# Copyright 2015 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Test class for common methods used by iRMC modules.
"""

import os
from unittest import mock

from oslo_config import cfg
from oslo_utils import uuidutils

from ironic.common import exception
from ironic.common import utils
from ironic.conductor import task_manager
from ironic.drivers.modules.irmc import common as irmc_common
from ironic.drivers.modules import snmp
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils
from ironic.tests.unit.drivers import third_party_driver_mock_specs \
    as mock_specs
from ironic.tests.unit.objects import utils as obj_utils


class BaseIRMCTest(db_base.DbTestCase):

    boot_interface = 'irmc-pxe'
    inspect_interface = 'irmc'
    power_interface = 'irmc'

    def setUp(self):
        super(BaseIRMCTest, self).setUp()
        self.config(enabled_hardware_types=['irmc', 'fake-hardware'],
                    enabled_power_interfaces=['irmc', 'fake'],
                    enabled_management_interfaces=['irmc', 'fake'],
                    enabled_bios_interfaces=['irmc', 'no-bios', 'fake'],
                    enabled_boot_interfaces=[self.boot_interface, 'fake'],
                    enabled_inspect_interfaces=['irmc', 'no-inspect', 'fake'])
        self.info = db_utils.get_test_irmc_info()
        self.node = obj_utils.create_test_node(
            self.context,
            driver='irmc',
            boot_interface=self.boot_interface,
            inspect_interface=self.inspect_interface,
            power_interface=self.power_interface,
            driver_info=self.info,
            uuid=uuidutils.generate_uuid())


class IRMCValidateParametersTestCase(BaseIRMCTest):

    @mock.patch.object(utils, 'is_fips_enabled',
                       return_value=False, autospec=True)
    def test_parse_driver_info(self, mock_check_fips):
        info = irmc_common.parse_driver_info(self.node)

        self.assertEqual('1.2.3.4', info['irmc_address'])
        self.assertEqual('admin0', info['irmc_username'])
        self.assertEqual('fake0', info['irmc_password'])
        self.assertEqual(60, info['irmc_client_timeout'])
        self.assertEqual(80, info['irmc_port'])
        self.assertEqual('digest', info['irmc_auth_method'])
        self.assertEqual('ipmitool', info['irmc_sensor_method'])
        self.assertEqual(snmp.SNMP_V2C, info['irmc_snmp_version'])
        self.assertEqual(161, info['irmc_snmp_port'])
        self.assertEqual('public', info['irmc_snmp_community'])
        self.assertTrue(info['irmc_verify_ca'])

    @mock.patch.object(utils, 'is_fips_enabled',
                       return_value=False, autospec=True)
    def test_parse_snmp_driver_info_with_snmp(self, mock_check_fips):
        test_list = [{'interfaces': [{'interface': 'inspect_interface',
                                      'impl': 'irmc'},
                                     {'interface': 'power_interface',
                                      'impl': 'irmc'}],
                      'snmp': True},
                     {'interfaces': [{'interface': 'inspect_interface',
                                      'impl': 'inspector'},
                                     {'interface': 'power_interface',
                                      'impl': 'irmc'}],
                      'snmp': True},
                     {'interfaces': [{'interface': 'inspect_interface',
                                      'impl': 'irmc'},
                                     {'interface': 'power_interface',
                                      'impl': 'ipmitool'}],
                      'snmp': True},
                     {'interfaces': [{'interface': 'inspect_interface',
                                      'impl': 'inspector'},
                                     {'interface': 'power_interface',
                                      'impl': 'ipmitool'}],
                      'snmp': False}
                     ]

        for t_conf in test_list:
            with self.subTest(t_conf=t_conf):
                for int_conf in t_conf['interfaces']:
                    setattr(self.node, int_conf['interface'], int_conf['impl'])
                irmc_common.parse_driver_info(self.node)

                if t_conf['snmp']:
                    mock_check_fips.assert_called()
                else:
                    mock_check_fips.assert_not_called()

                mock_check_fips.reset_mock()

    def test_parse_driver_info_snmpv3(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3'
        self.node.driver_info['irmc_snmp_user'] = 'admin0'
        self.node.driver_info['irmc_snmp_auth_password'] = 'valid_key'
        self.node.driver_info['irmc_snmp_priv_password'] = 'valid_key'
        info = irmc_common.parse_driver_info(self.node)

        self.assertEqual('1.2.3.4', info['irmc_address'])
        self.assertEqual('admin0', info['irmc_username'])
        self.assertEqual('fake0', info['irmc_password'])
        self.assertEqual(60, info['irmc_client_timeout'])
        self.assertEqual(80, info['irmc_port'])
        self.assertEqual('digest', info['irmc_auth_method'])
        self.assertEqual('ipmitool', info['irmc_sensor_method'])
        self.assertEqual(snmp.SNMP_V3, info['irmc_snmp_version'])
        self.assertEqual(161, info['irmc_snmp_port'])
        self.assertEqual('public', info['irmc_snmp_community'])
        self.assertEqual('admin0', info['irmc_snmp_user'])
        self.assertEqual(snmp.snmp_auth_protocols['sha'],
                         info['irmc_snmp_auth_proto'])
        self.assertEqual('valid_key', info['irmc_snmp_auth_password'])
        self.assertEqual(snmp.snmp_priv_protocols['aes'],
                         info['irmc_snmp_priv_proto'])
        self.assertEqual('valid_key', info['irmc_snmp_priv_password'])

    @mock.patch.object(utils, 'is_fips_enabled',
                       return_value=False, autospec=True)
    def test_parse_driver_option_default(self, mock_check_fips):
        self.node.driver_info = {
            "irmc_address": "1.2.3.4",
            "irmc_username": "admin0",
            "irmc_password": "fake0",
        }
        info = irmc_common.parse_driver_info(self.node)

        self.assertEqual('basic', info['irmc_auth_method'])
        self.assertEqual(443, info['irmc_port'])
        self.assertEqual(60, info['irmc_client_timeout'])
        self.assertEqual('ipmitool', info['irmc_sensor_method'])
        self.assertEqual(True, info['irmc_verify_ca'])

    def test_parse_driver_info_missing_address(self):
        del self.node.driver_info['irmc_address']
        self.assertRaises(exception.MissingParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_missing_username(self):
        del self.node.driver_info['irmc_username']
        self.assertRaises(exception.MissingParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_missing_password(self):
        del self.node.driver_info['irmc_password']
        self.assertRaises(exception.MissingParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_invalid_timeout(self):
        self.node.driver_info['irmc_client_timeout'] = 'qwe'
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_invalid_port(self):
        self.node.driver_info['irmc_port'] = 'qwe'
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_invalid_auth_method(self):
        self.node.driver_info['irmc_auth_method'] = 'qwe'
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_invalid_sensor_method(self):
        self.node.driver_info['irmc_sensor_method'] = 'qwe'
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_missing_multiple_params(self):
        del self.node.driver_info['irmc_password']
        del self.node.driver_info['irmc_address']
        e = self.assertRaises(exception.MissingParameterValue,
                              irmc_common.parse_driver_info, self.node)
        self.assertIn('irmc_password', str(e))
        self.assertIn('irmc_address', str(e))

    def test_parse_driver_info_invalid_snmp_version(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3x'
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    @mock.patch.object(utils, 'is_fips_enabled',
                       return_value=True, autospec=True)
    def test_parse_driver_info_invalid_snmp_version_fips(self,
                                                         mock_check_fips):
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)
        self.assertEqual(1, mock_check_fips.call_count)

    def test_parse_driver_info_invalid_snmp_port(self):
        self.node.driver_info['irmc_snmp_port'] = '161p'
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_invalid_snmp_community(self):
        self.node.driver_info['irmc_snmp_version'] = 'v2c'
        self.node.driver_info['irmc_snmp_community'] = 100
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_missing_snmp_user(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3'
        self.node.driver_info['irmc_snmp_auth_password'] = 'valid_key'
        self.node.driver_info['irmc_snmp_priv_password'] = 'valid_key'
        self.assertRaises(exception.MissingParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_missing_snmp_auth_password(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3'
        self.node.driver_info['irmc_snmp_user'] = 'admin0'
        self.node.driver_info['irmc_snmp_priv_password'] = 'valid_key'
        self.assertRaises(exception.MissingParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_missing_snmp_priv_password(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3'
        self.node.driver_info['irmc_snmp_user'] = 'admin0'
        self.node.driver_info['irmc_snmp_auth_password'] = 'valid_key'
        self.assertRaises(exception.MissingParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_using_snmp_security(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3'
        self.node.driver_info['irmc_snmp_security'] = 'admin0'
        self.node.driver_info['irmc_snmp_auth_password'] = 'valid_key'
        self.node.driver_info['irmc_snmp_priv_password'] = 'valid_key'
        info = irmc_common.parse_driver_info(self.node)
        self.assertEqual('admin0', info['irmc_snmp_user'])

    def test_parse_driver_info_invalid_snmp_security(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3'
        self.node.driver_info['irmc_snmp_security'] = 100
        self.node.driver_info['irmc_snmp_auth_password'] = 'valid_key'
        self.node.driver_info['irmc_snmp_priv_password'] = 'valid_key'
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_invalid_snmp_user(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3'
        self.node.driver_info['irmc_snmp_user'] = 100
        self.node.driver_info['irmc_snmp_auth_password'] = 'valid_key'
        self.node.driver_info['irmc_snmp_priv_password'] = 'valid_key'
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_invalid_snmp_auth_password(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3'
        self.node.driver_info['irmc_snmp_user'] = 'admin0'
        self.node.driver_info['irmc_snmp_auth_password'] = 100
        self.node.driver_info['irmc_snmp_priv_password'] = 'valid_key'
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_short_snmp_auth_password(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3'
        self.node.driver_info['irmc_snmp_user'] = 'admin0'
        self.node.driver_info['irmc_snmp_auth_password'] = 'short'
        self.node.driver_info['irmc_snmp_priv_password'] = 'valid_key'
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_invalid_snmp_priv_password(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3'
        self.node.driver_info['irmc_snmp_user'] = 'admin0'
        self.node.driver_info['irmc_snmp_auth_password'] = 'valid_key'
        self.node.driver_info['irmc_snmp_priv_password'] = 100
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_short_snmp_priv_password(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3'
        self.node.driver_info['irmc_snmp_user'] = 'admin0'
        self.node.driver_info['irmc_snmp_auth_password'] = 'valid_key'
        self.node.driver_info['irmc_snmp_priv_password'] = 'short'
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_invalid_snmp_auth_proto(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3'
        self.node.driver_info['irmc_snmp_user'] = 'admin0'
        self.node.driver_info['irmc_snmp_auth_password'] = 'valid_key'
        self.node.driver_info['irmc_snmp_priv_password'] = 'valid_key'
        self.node.driver_info['irmc_snmp_auth_proto'] = 'invalid'
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    def test_parse_driver_info_invalid_snmp_priv_proto(self):
        self.node.driver_info['irmc_snmp_version'] = 'v3'
        self.node.driver_info['irmc_snmp_user'] = 'admin0'
        self.node.driver_info['irmc_snmp_auth_password'] = 'valid_key'
        self.node.driver_info['irmc_snmp_priv_password'] = 'valid_key'
        self.node.driver_info['irmc_snmp_priv_proto'] = 'invalid'
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)

    @mock.patch.object(os.path, 'isabs', return_value=True, autospec=True)
    @mock.patch.object(os.path, 'isdir', return_value=True, autospec=True)
    def test_parse_driver_info_dir_path_verify_ca(self, mock_isdir,
                                                  mock_isabs):
        fake_path = 'absolute/path/to/a/valid/CA'
        self.node.driver_info['irmc_verify_ca'] = fake_path
        info = irmc_common.parse_driver_info(self.node)
        self.assertEqual(fake_path, info['irmc_verify_ca'])
        mock_isdir.assert_called_once_with(fake_path)
        mock_isabs.assert_called_once_with(fake_path)

    @mock.patch.object(os.path, 'isabs', return_value=True, autospec=True)
    @mock.patch.object(os.path, 'isfile', return_value=True, autospec=True)
    def test_parse_driver_info_file_path_verify_ca(self, mock_isfile,
                                                   mock_isabs):
        fake_path = 'absolute/path/to/a/valid/ca.pem'
        self.node.driver_info['irmc_verify_ca'] = fake_path
        info = irmc_common.parse_driver_info(self.node)
        self.assertEqual(fake_path, info['irmc_verify_ca'])
        mock_isfile.assert_called_once_with(fake_path)
        mock_isabs.assert_called_once_with(fake_path)

    def test_parse_driver_info_string_bool_verify_ca(self):
        self.node.driver_info['irmc_verify_ca'] = "False"
        info = irmc_common.parse_driver_info(self.node)
        self.assertFalse(info['irmc_verify_ca'])

    def test_parse_driver_info_invalid_verify_ca(self):
        self.node.driver_info['irmc_verify_ca'] = "1234"
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)
        self.node.driver_info['irmc_verify_ca'] = 1234
        self.assertRaises(exception.InvalidParameterValue,
                          irmc_common.parse_driver_info, self.node)


class IRMCCommonMethodsTestCase(BaseIRMCTest):

    @mock.patch.object(irmc_common, 'scci',
                       spec_set=mock_specs.SCCICLIENT_IRMC_SCCI_SPEC)
    def test_get_irmc_client(self, mock_scci):
        self.info['irmc_port'] = 80
        self.info['irmc_auth_method'] = 'digest'
        self.info['irmc_client_timeout'] = 60
        self.info['irmc_verify_ca'] = True
        mock_scci.get_client.return_value = 'get_client'
        returned_mock_scci_get_client = irmc_common.get_irmc_client(self.node)
        mock_scci.get_client.assert_called_with(
            self.info['irmc_address'],
            self.info['irmc_username'],
            self.info['irmc_password'],
            port=self.info['irmc_port'],
            auth_method=self.info['irmc_auth_method'],
            verify=self.info['irmc_verify_ca'],
            client_timeout=self.info['irmc_client_timeout'])
        self.assertEqual('get_client', returned_mock_scci_get_client)

    def test_update_ipmi_properties(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            ipmi_info = {
                "ipmi_address": "1.2.3.4",
                "ipmi_username": "admin0",
                "ipmi_password": "fake0",
            }
            task.node.driver_info = self.info
            irmc_common.update_ipmi_properties(task)
            actual_info = task.node.driver_info
            expected_info = dict(self.info, **ipmi_info)
            self.assertEqual(expected_info, actual_info)

    @mock.patch.object(irmc_common, 'scci',
                       spec_set=mock_specs.SCCICLIENT_IRMC_SCCI_SPEC)
    def test_get_irmc_report(self, mock_scci):
        self.info['irmc_port'] = 80
        self.info['irmc_auth_method'] = 'digest'
        self.info['irmc_client_timeout'] = 60
        self.info['irmc_verify_ca'] = True
        mock_scci.get_report.return_value = 'get_report'
        returned_mock_scci_get_report = irmc_common.get_irmc_report(self.node)
        mock_scci.get_report.assert_called_with(
            self.info['irmc_address'],
            self.info['irmc_username'],
            self.info['irmc_password'],
            port=self.info['irmc_port'],
            auth_method=self.info['irmc_auth_method'],
            verify=self.info['irmc_verify_ca'],
            client_timeout=self.info['irmc_client_timeout'])
        self.assertEqual('get_report', returned_mock_scci_get_report)

    def test_out_range_port(self):
        self.assertRaises(ValueError, cfg.CONF.set_override,
                          'port', 60, 'irmc')

    def test_out_range_auth_method(self):
        self.assertRaises(ValueError, cfg.CONF.set_override,
                          'auth_method', 'fake', 'irmc')

    def test_out_range_sensor_method(self):
        self.assertRaises(ValueError, cfg.CONF.set_override,
                          'sensor_method', 'fake', 'irmc')

    @mock.patch.object(irmc_common, 'elcm',
                       spec_set=mock_specs.SCCICLIENT_IRMC_ELCM_SPEC)
    def test_set_secure_boot_mode_enable(self, mock_elcm):
        mock_elcm.set_secure_boot_mode.return_value = 'set_secure_boot_mode'
        info = irmc_common.parse_driver_info(self.node)
        irmc_common.set_secure_boot_mode(self.node, True)
        mock_elcm.set_secure_boot_mode.assert_called_once_with(
            info, True)

    @mock.patch.object(irmc_common, 'elcm',
                       spec_set=mock_specs.SCCICLIENT_IRMC_ELCM_SPEC)
    def test_set_secure_boot_mode_disable(self, mock_elcm):
        mock_elcm.set_secure_boot_mode.return_value = 'set_secure_boot_mode'
        info = irmc_common.parse_driver_info(self.node)
        irmc_common.set_secure_boot_mode(self.node, False)
        mock_elcm.set_secure_boot_mode.assert_called_once_with(
            info, False)

    @mock.patch.object(irmc_common, 'elcm',
                       spec_set=mock_specs.SCCICLIENT_IRMC_ELCM_SPEC)
    @mock.patch.object(irmc_common, 'scci',
                       spec_set=mock_specs.SCCICLIENT_IRMC_SCCI_SPEC)
    def test_set_secure_boot_mode_fail(self, mock_scci, mock_elcm):
        irmc_common.scci.SCCIError = Exception
        mock_elcm.set_secure_boot_mode.side_effect = Exception
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.assertRaises(exception.IRMCOperationError,
                              irmc_common.set_secure_boot_mode,
                              task.node, True)
            info = irmc_common.parse_driver_info(task.node)
            mock_elcm.set_secure_boot_mode.assert_called_once_with(
                info, True)

    @mock.patch.object(irmc_common, 'elcm',
                       spec_set=mock_specs.SCCICLIENT_IRMC_ELCM_SPEC)
    def test_check_elcm_license_success_with_200(self, elcm_mock):
        elcm_req_mock = elcm_mock.elcm_request
        json_data = ('{ "eLCMStatus" : { "EnabledAndLicenced" : "true" , '
                     '"SDCardMounted" : "false" } }')
        func_return_value = {'active': True, 'status_code': 200}
        response_mock = elcm_req_mock.return_value
        response_mock.status_code = 200
        response_mock.text = json_data
        self.assertEqual(irmc_common.check_elcm_license(self.node),
                         func_return_value)

    @mock.patch.object(irmc_common, 'elcm',
                       spec_set=mock_specs.SCCICLIENT_IRMC_ELCM_SPEC)
    def test_check_elcm_license_success_with_500(self, elcm_mock):
        elcm_req_mock = elcm_mock.elcm_request
        json_data = ''
        func_return_value = {'active': False, 'status_code': 500}
        response_mock = elcm_req_mock.return_value
        response_mock.status_code = 500
        response_mock.text = json_data
        self.assertEqual(irmc_common.check_elcm_license(self.node),
                         func_return_value)

    @mock.patch.object(irmc_common, 'scci',
                       spec_set=mock_specs.SCCICLIENT_IRMC_SCCI_SPEC)
    @mock.patch.object(irmc_common, 'elcm',
                       spec_set=mock_specs.SCCICLIENT_IRMC_ELCM_SPEC)
    def test_check_elcm_license_fail_invalid_json(self, elcm_mock, scci_mock):
        scci_mock.SCCIError = Exception
        elcm_req_mock = elcm_mock.elcm_request
        json_data = ''
        response_mock = elcm_req_mock.return_value
        response_mock.status_code = 200
        response_mock.text = json_data
        self.assertRaises(exception.IRMCOperationError,
                          irmc_common.check_elcm_license, self.node)

    @mock.patch.object(irmc_common, 'scci',
                       spec_set=mock_specs.SCCICLIENT_IRMC_SCCI_SPEC)
    @mock.patch.object(irmc_common, 'elcm',
                       spec_set=mock_specs.SCCICLIENT_IRMC_ELCM_SPEC)
    def test_check_elcm_license_fail_elcm_error(self, elcm_mock, scci_mock):
        scci_mock.SCCIError = Exception
        elcm_req_mock = elcm_mock.elcm_request
        elcm_req_mock.side_effect = scci_mock.SCCIError
        self.assertRaises(exception.IRMCOperationError,
                          irmc_common.check_elcm_license, self.node)

    @mock.patch.object(irmc_common, 'get_irmc_report', autospec=True)
    @mock.patch.object(irmc_common, 'scci',
                       spec_set=mock_specs.SCCICLIENT_IRMC_SCCI_SPEC)
    def test_set_irmc_version_success(self, scci_mock, get_report_mock):
        version_str = 'iRMC S6/2.00'
        scci_mock.get_irmc_version_str.return_value = version_str.split('/')
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            irmc_common.set_irmc_version(task)
            self.assertEqual(version_str,
                             task.node.driver_internal_info['irmc_fw_version'])

    @mock.patch.object(irmc_common, 'get_irmc_report', autospec=True)
    @mock.patch.object(irmc_common, 'scci',
                       spec_set=mock_specs.SCCICLIENT_IRMC_SCCI_SPEC)
    def test_set_irmc_version_fail(self, scci_mock, get_report_mock):
        scci_mock.SCCIError = Exception
        get_report_mock.side_effect = scci_mock.SCCIError
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            self.assertRaises(exception.IRMCOperationError,
                              irmc_common.set_irmc_version, task)

    def test_within_version_ranges_success(self):
        self.node.set_driver_internal_info('irmc_fw_version', 'iRMC S6/2.00')
        ver_range_list = [
            {'4': {'upper': '1.05'},
             '6': {'min': '1.95', 'upper': '2.01'}
             },
            {'4': {'upper': '1.05'},
             '6': {'min': '1.95', 'upper': None}
             },
            {'4': {'upper': '1.05'},
             '6': {'min': '1.95'}
             },
            {'4': {'upper': '1.05'},
             '6': {}
             },
            {'4': {'upper': '1.05'},
             '6': None
             }]
        for range_dict in ver_range_list:
            with self.subTest():
                self.assertTrue(irmc_common.within_version_ranges(self.node,
                                                                  range_dict))

    def test_within_version_ranges_success_out_range(self):
        self.node.set_driver_internal_info('irmc_fw_version', 'iRMC S6/2.00')
        ver_range_list = [
            {'4': {'upper': '1.05'},
             '6': {'min': '1.95', 'upper': '2.00'}
             },
            {'4': {'upper': '1.05'},
             '6': {'min': '1.95', 'upper': '1.99'}
             },
            {'4': {'upper': '1.05'},
             }]
        for range_dict in ver_range_list:
            with self.subTest():
                self.assertFalse(irmc_common.within_version_ranges(self.node,
                                                                   range_dict))

    def test_within_version_ranges_fail_no_match(self):
        self.node.set_driver_internal_info('irmc_fw_version', 'ver/2.00')
        ver_range = {
            '4': {'upper': '1.05'},
            '6': {'min': '1.95', 'upper': '2.01'}
        }
        self.assertFalse(irmc_common.within_version_ranges(self.node,
                                                           ver_range))

    def test_within_version_ranges_fail_no_version_set(self):
        ver_range = {
            '4': {'upper': '1.05'},
            '6': {'min': '1.95', 'upper': '2.01'}
        }
        self.assertFalse(irmc_common.within_version_ranges(self.node,
                                                           ver_range))



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
DRAC vendor-passthru interface
"""

from ironic.drivers.modules.redfish import vendor as redfish_vendor


class DracRedfishVendorPassthru(redfish_vendor.RedfishVendorPassthru):
    """iDRAC Redfish interface for vendor_passthru.

    Use the Redfish implementation for vendor passthru.
    """
    # NOTE(cardoe): deprecated in favor of plain Redfish
    supported = False



# Copyright 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# NOTE(tenbrae): since __init__ is loaded before the files in the same
#             directory, and some third-party driver tests may need to have
#             their external libraries mocked, we load the file which does
#             that mocking here -- in the __init__.

from ironic.tests.unit.drivers import third_party_driver_mocks   # noqa



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hashlib
import string

import requests

# adapted from IPA
DEFAULT_CHUNK_SIZE = 1024 * 1024  # 1MB


class StreamingDownloader(object):

    def __init__(self, url, chunksize, hash_algo=None, verify=True,
                 certs=None):
        if hash_algo is not None:
            self.hasher = hashlib.new(hash_algo)
        else:
            self.hasher = None
        self.chunksize = chunksize
        resp = requests.get(url, stream=True, verify=verify, cert=certs,
                            timeout=30)
        if resp.status_code != 200:
            raise Exception('Invalid response code: %s' % resp.status_code)

        self._request = resp

    def __iter__(self):
        for chunk in self._request.iter_content(chunk_size=self.chunksize):
            if self.hasher is not None:
                self.hasher.update(chunk)
            yield chunk

    def checksum(self):
        if self.hasher is not None:
            return self.hasher.hexdigest()


def stream_to_dest(url, dest, chunksize, hash_algo, verify=True, certs=None):
    downloader = StreamingDownloader(url, chunksize, hash_algo,
                                     verify=verify, certs=certs)

    with open(dest, 'wb+') as f:
        for chunk in downloader:
            f.write(chunk)

    return downloader.checksum()


def main():
    module = AnsibleModule(  # noqa This is normal for Ansible modules.
        argument_spec=dict(
            url=dict(required=True, type='str'),
            dest=dict(required=True, type='str'),
            checksum=dict(required=False, type='str', default=''),
            chunksize=dict(required=False, type='int',
                           default=DEFAULT_CHUNK_SIZE),
            validate_certs=dict(required=False, type='bool', default=True),
            client_cert=dict(required=False, type='str', default=''),
            client_key=dict(required=False, type='str', default='')

        ))

    url = module.params['url']
    dest = module.params['dest']
    checksum = module.params['checksum']
    chunksize = module.params['chunksize']
    validate = module.params['validate_certs']
    client_cert = module.params['client_cert']
    client_key = module.params['client_key']
    if client_cert:
        certs = (client_cert, client_key) if client_key else client_cert
    else:
        certs = None

    if checksum == '':
        hash_algo, checksum = None, None
    else:
        try:
            hash_algo, checksum = checksum.rsplit(':', 1)
        except ValueError:
            module.fail_json(msg='The checksum parameter has to be in format '
                             '"<algorithm>:<checksum>"')
        checksum = checksum.lower()
        if not all(c in string.hexdigits for c in checksum):
            module.fail_json(msg='The checksum must be valid HEX number')

        if hash_algo not in hashlib.algorithms_available:
            module.fail_json(msg="%s checksums are not supported" % hash_algo)

    try:
        actual_checksum = stream_to_dest(
            url, dest, chunksize, hash_algo, verify=validate, certs=certs)
    except Exception as e:
        module.fail_json(msg=str(e))
    else:
        if hash_algo and actual_checksum != checksum:
            module.fail_json(msg='Invalid dest checksum')
        else:
            module.exit_json(changed=True)


# NOTE(pas-ha) Ansible's module_utils.basic is licensed under BSD (2 clause)
from ansible.module_utils.basic import *  # noqa
if __name__ == '__main__':
    main()



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic import api
from ironic.api.controllers import link


def has_next(collection, limit):
    """Return whether collection has more items."""
    return len(collection) and len(collection) == limit


def list_convert_with_links(items, item_name, limit, url, fields=None,
                            sanitize_func=None, key_field='uuid',
                            sanitizer_args=None, **kwargs):
    """Build a collection dict including the next link for paging support.

    :param items:
        List of unsanitized items to include in the collection
    :param item_name:
        Name of dict key for items value
    :param limit:
        Paging limit
    :param url:
        Base URL for building next link
    :param fields:
        Optional fields to use for sanitize function
    :param sanitize_func:
        Optional sanitize function run on each item, item changes will be
        done in-place
    :param key_field:
        Key name for building next URL
    :param sanitizer_args:
        Dictionary with additional arguments to be passed to the sanitizer.
    :param kwargs:
        other arguments passed to ``get_next``
    :returns:
        A dict containing ``item_name`` and ``next`` values
    """
    assert url, "BUG: collections require a base URL"
    assert limit is None or isinstance(limit, int), \
        f"BUG: limit must be None or int, got {type(limit)}"

    items_dict = {
        item_name: items
    }
    next_uuid = get_next(
        items, limit, url=url, fields=fields, key_field=key_field, **kwargs)
    if next_uuid:
        items_dict['next'] = next_uuid

    if sanitize_func:
        if sanitizer_args:
            for item in items:
                sanitize_func(item, fields, **sanitizer_args)
        else:
            for item in items:
                sanitize_func(item, fields=fields)

    return items_dict


def get_next(collection, limit, url, key_field='uuid', **kwargs):
    """Return a link to the next subset of the collection."""
    if not has_next(collection, limit):
        return None

    fields = kwargs.pop('fields', None)
    # NOTE(saga): If fields argument is present in kwargs and not None. It
    # is a list so convert it into a comma separated string.
    if fields:
        kwargs['fields'] = ','.join(fields)
    q_args = ''.join(['%s=%s&' % (key, kwargs[key]) for key in kwargs])

    last_item = collection[-1]
    # handle items which are either objects or dicts
    if hasattr(last_item, key_field):
        marker = getattr(last_item, key_field)
    else:
        marker = last_item.get(key_field)

    next_args = '?%(args)slimit=%(limit)d&marker=%(marker)s' % {
        'args': q_args, 'limit': limit,
        'marker': marker}

    return link.make_link('next', api.request.public_url,
                          url, next_args)['href']



# Copyright (c) 2024 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import os
import re
import time
from urllib import parse as urlparse

from oslo_log import log as logging
from oslo_utils import fileutils

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import image_service
from ironic.conf import CONF

LOG = logging.getLogger(__name__)


# REGEX matches for Checksum file payloads
# If this list requires changes, it should be changed in
# ironic-python-agent (extensions/standby.py) as well.

MD5_MATCH = r"^([a-fA-F\d]{32})\s"  # MD5 at beginning of line
MD5_MATCH_END = r"\s([a-fA-F\d]{32})$"  # MD5 at end of line
MD5_MATCH_ONLY = r"^([a-fA-F\d]{32})$"  # MD5 only
SHA256_MATCH = r"^([a-fA-F\d]{64})\s"  # SHA256 at beginning of line
SHA256_MATCH_END = r"\s([a-fA-F\d]{64})$"  # SHA256 at end of line
SHA256_MATCH_ONLY = r"^([a-fA-F\d]{64})$"  # SHA256 only
SHA512_MATCH = r"^([a-fA-F\d]{128})\s"  # SHA512 at beginning of line
SHA512_MATCH_END = r"\s([a-fA-F\d]{128})$"  # SHA512 at end of line
SHA512_MATCH_ONLY = r"^([a-fA-F\d]{128})$"  # SHA512 only
FILENAME_MATCH_END = r"\s[*]?{filename}$"  # Filename binary/text end of line
FILENAME_MATCH_PARENTHESES = r"\s\({filename}\)\s"  # CentOS images

CHECKSUM_MATCHERS = (MD5_MATCH, MD5_MATCH_END, SHA256_MATCH, SHA256_MATCH_END,
                     SHA512_MATCH, SHA512_MATCH_END)
CHECKSUM_ONLY_MATCHERS = (MD5_MATCH_ONLY, SHA256_MATCH_ONLY, SHA512_MATCH_ONLY)
FILENAME_MATCHERS = (FILENAME_MATCH_END, FILENAME_MATCH_PARENTHESES)


def validate_checksum(path, checksum, checksum_algo=None):
    """Validate image checksum.

    :param path: File path in the form of a string to calculate a checksum
                 which is compared to the checksum field.
    :param checksum: The supplied checksum value, a string, which will be
                     compared to the file.
    :param checksum_algo: The checksum type of the algorithm.
    :raises: ImageChecksumError if the supplied data cannot be parsed or
             if the supplied value does not match the supplied checksum
             value.
    """
    # TODO(TheJilia): At some point, we likely need to compare
    # the incoming checksum algorithm upfront, ut if one is invoked which
    # is not supported, hashlib will raise ValueError.
    use_checksum_algo = None
    if ":" in checksum:
        # A form of communicating the checksum algorithm is to delimit the
        # type from the value. See ansible deploy interface where this
        # is most evident.
        split_checksum = checksum.split(":")
        use_checksum = split_checksum[1]
        use_checksum_algo = split_checksum[0]
    else:
        use_checksum = checksum
    if not use_checksum_algo:
        use_checksum_algo = checksum_algo
    # If we have a zero length value, but we split it, we have
    # invalid input. Also, checksum is what we expect, algorithm is
    # optional. This guards against the split of a value which is
    # image_checksum = "sha256:" which is a potential side effect of
    # splitting the string.
    if use_checksum == '':
        raise exception.ImageChecksumError()

    # Make everything lower case since we don't expect mixed case,
    # but we may have human originated input on the supplied algorithm.
    try:
        if not use_checksum_algo:
            # This is backwards compatible support for a bare checksum.
            calculated = compute_image_checksum(path)
        else:
            calculated = compute_image_checksum(path,
                                                use_checksum_algo.lower())
    except ValueError:
        # ValueError is raised when an invalid/unsupported/unknown
        # checksum algorithm is invoked.
        LOG.error("Failed to generate checksum for file %(path)s, possible "
                  "invalid checksum algorithm: %(algo)s",
                  {"path": path,
                   "algo": use_checksum_algo})
        raise exception.ImageChecksumAlgorithmFailure()
    except OSError:
        LOG.error("Failed to read file %(path)s to compute checksum.",
                  {"path": path})
        raise exception.ImageChecksumFileReadFailure()
    if (use_checksum is not None
        and calculated.lower() != use_checksum.lower()):
        LOG.error("We were supplied a checksum value of %(supplied)s, but "
                  "calculated a value of %(value)s. This is a fatal error.",
                  {"supplied": use_checksum,
                   "value": calculated})
        raise exception.ImageChecksumError()


def compute_image_checksum(image_path, algorithm='md5'):
    """Compute checksum by given image path and algorithm.

    :param image_path: The path to the file to undergo checksum calculation.
    :param algorithm: The checksum algorithm to utilize. Defaults
        to 'md5' due to historical support reasons in Ironic.
    :returns: The calculated checksum value.
    :raises: ValueError when the checksum algorithm is not supported
       by the system.
    """

    time_start = time.time()
    LOG.debug('Start computing %(algo)s checksum for image %(image)s.',
              {'algo': algorithm, 'image': image_path})

    checksum = fileutils.compute_file_checksum(image_path,
                                               algorithm=algorithm)
    time_elapsed = time.time() - time_start
    LOG.debug('Computed %(algo)s checksum for image %(image)s in '
              '%(delta).2f seconds, checksum value: %(checksum)s.',
              {'algo': algorithm, 'image': image_path, 'delta': time_elapsed,
               'checksum': checksum})
    return checksum


def get_checksum_and_algo(instance_info):
    """Get and return the image checksum and algo.

    :param instance_info: The node instance info, or newly updated/generated
                          instance_info value.
    :returns: A tuple containing two values, a checksum and algorithm,
              if available.
    """
    checksum_algo = None
    if 'image_os_hash_value' in instance_info.keys():
        # A value set by image_os_hash_value supersedes other
        # possible uses as it is specific.
        checksum = instance_info.get('image_os_hash_value')
        checksum_algo = instance_info.get('image_os_hash_algo')
    else:
        checksum = instance_info.get('image_checksum')
        image_source = instance_info.get('image_source')

        # NOTE(stevebaker): file:// images have no requirement to supply
        # checksums but they are now mandatory for validation as part
        # of the fix for CVE-2024-47211.
        # The only practical option is to calculate it here.
        if checksum is None and image_source.startswith('file:'):
            checksum_algo = "sha256"
            image_path = urlparse.urlparse(image_source).path
            checksum = fileutils.compute_file_checksum(
                image_path, algorithm=checksum_algo)

        elif is_checksum_url(checksum):
            checksum = get_checksum_from_url(checksum, image_source)

        # NOTE(TheJulia): This is all based on SHA-2 lengths.
        # SHA-3 would require a hint and it would not be a fixed length.
        # That said, SHA-2 is still valid and has not been withdrawn.
        checksum_len = len(checksum)
        if checksum_len == 128:
            # SHA2-512 is 512 bits, 128 characters.
            checksum_algo = "sha512"
        elif checksum_len == 64:
            checksum_algo = "sha256"

        if checksum_len == 32 and not CONF.agent.allow_md5_checksum:
            # MD5 not permitted and the checksum is the length of MD5
            # and not otherwise defined.
            LOG.error('Cannot compute the checksum as it uses MD5 '
                      'and is disabled by configuration. If the checksum '
                      'is *not* MD5, please specify the algorithm.')
            raise exception.ImageChecksumAlgorithmFailure()

    return checksum, checksum_algo


def is_checksum_url(checksum):
    """Identify if checksum is not a url.

    :param checksum: The user supplied checksum value.
    :returns: True if the checksum is a url, otherwise False.
    :raises: ImageChecksumURLNotSupported should the conductor have this
             support disabled.
    """
    if (checksum.startswith('http://') or checksum.startswith('https://')):
        if CONF.conductor.disable_support_for_checksum_files:
            raise exception.ImageChecksumURLNotSupported()
        return True
    else:
        return False


def get_checksum_from_url(checksum, image_source):
    """Gets a checksum value based upon a remote checksum URL file.

    :param checksum: The URL to the checksum URL content.
    :param image_soource: The image source utilized to match with
        the contents of the URL payload file.
    :raises: ImageDownloadFailed when the checksum file cannot be
        accessed or cannot be parsed.
    """

    LOG.debug('Attempting to download checksum from: %(checksum)s.',
              {'checksum': checksum})

    # Directly invoke the image service and get the checksum data.
    resp = image_service.HttpImageService.get(checksum)
    checksum_url = str(checksum)

    # NOTE(TheJulia): The rest of this method is taken from
    # ironic-python-agent. If a change is required here, it may
    # be required in ironic-python-agent (extensions/standby.py).
    lines = [line.strip() for line in resp.split('\n') if line.strip()]
    if not lines:
        raise exception.ImageDownloadFailed(image_href=checksum,
                                            reason=_('Checksum file empty.'))
    elif len(lines) == 1:
        # Special case - checksums file with only the checksum itself
        if ' ' not in lines[0]:
            for matcher in CHECKSUM_ONLY_MATCHERS:
                checksum = re.findall(matcher, lines[0])
                if checksum:
                    return checksum[0]
            raise exception.ImageDownloadFailed(
                image_href=checksum_url,
                reason=(
                    _("Invalid checksum file (No valid checksum found)")))
    # FIXME(dtantsur): can we assume the same name for all images?
    expected_fname = os.path.basename(urlparse.urlparse(
        image_source).path)
    for line in lines:
        # Ignore comment lines
        if line.startswith("#"):
            continue

        # Ignore checksums for other files
        for matcher in FILENAME_MATCHERS:
            if re.findall(matcher.format(filename=expected_fname), line):
                break
        else:
            continue

        for matcher in CHECKSUM_MATCHERS:
            checksum = re.findall(matcher, line)
            if checksum:
                return checksum[0]

    raise exception.ImageDownloadFailed(
        image_href=checksum,
        reason=(_("Checksum file does not contain name %s")
                % expected_fname))


class TransferHelper(object):

    def __init__(self, response, checksum_algo, expected_checksum):
        """Helper class to drive data download with concurrent checksum.

        The TransferHelper can be used to help retrieve data from a
        Python requests request invocation, where the request was set
        with `stream=True`, which also builds the checksum digest as the
        transfer is underway.

        :param response: A populated requests.model.Response object.
        :param checksum_algo: The expected checksum algorithm.
        :param expected_checksum: The expected checksum of the data being
                                  transferred.

        """
        # NOTE(TheJulia): Similar code exists in IPA in regards to
        # downloading and checksumming a raw image while streaming.
        # If a change is required here, it might be worthwhile to
        # consider if a similar change is needed in IPA.
        # NOTE(TheJulia): 1 Megabyte is an attempt to always exceed the
        # minimum chunk size which may be needed for proper checksum
        # generation and balance the memory required. We may want to
        # tune this, but 1MB has worked quite well for IPA for some time.
        # This may artificially throttle transfer speeds a little in
        # high performance environments as the data may get held up
        # in the kernel limiting the window from scaling.
        self._chunk_size = 1024 * 1024  # 1MB
        self._last_check_time = time.time()
        self._request = response
        self._bytes_transferred = 0
        self._checksum_algo = checksum_algo
        self._expected_checksum = expected_checksum
        self._expected_size = self._request.headers.get(
            'Content-Length')
        # Determine the hash algorithm and value will be used for calculation
        # and verification, fallback to md5 if algorithm is not set or not
        # supported.
        # NOTE(TheJulia): Regarding MD5, it is likely this will never be
        # hit, but we will guard in case of future use for this method
        # anyhow.
        if checksum_algo == 'md5' and not CONF.agent.allow_md5_checksum:
            # MD5 not permitted
            LOG.error('MD5 checksum utilization is disabled by '
                      'configuration.')
            raise exception.ImageChecksumAlgorithmFailure()

        if checksum_algo in hashlib.algorithms_available:
            self._hash_algo = hashlib.new(checksum_algo)
        else:
            raise ValueError("Unable to process checksum processing "
                             "for image transfer. Algorithm %s "
                             "is not available." % checksum_algo)

    def __iter__(self):
        """Downloads and returns the next chunk of the image.

        :returns: A chunk of the image. Size of 1MB.
        """
        self._last_chunk_time = None
        for chunk in self._request.iter_content(self._chunk_size):
            # Per requests forum posts/discussions, iter_content should
            # periodically yield to the caller for the client to do things
            # like stopwatch and potentially interrupt the download.
            # While this seems weird and doesn't exactly seem to match the
            # patterns in requests and urllib3, it does appear to be the
            # case. Field testing in environments where TCP sockets were
            # discovered in a read hanged state were navigated with
            # this code in IPA.
            if chunk:
                self._last_chunk_time = time.time()
                if isinstance(chunk, str):
                    encoded_data = chunk.encode()
                    self._hash_algo.update(encoded_data)
                    self._bytes_transferred += len(encoded_data)
                else:
                    self._hash_algo.update(chunk)
                    self._bytes_transferred += len(chunk)
                yield chunk
            elif (time.time() - self._last_chunk_time
                  > CONF.image_download_connection_timeout):
                LOG.error('Timeout reached waiting for a chunk of data from '
                          'a remote server.')
                raise exception.ImageDownloadError(
                    self._image_info['id'],
                    'Timed out reading next chunk from webserver')

    @property
    def checksum_matches(self):
        """Verifies the checksum matches and returns True/False."""
        checksum = self._hash_algo.hexdigest()
        if checksum != self._expected_checksum:
            # This is a property, let the caller figure out what it
            # wants to do.
            LOG.error('Verifying transfer checksum %(algo_name)s value '
                      '%(checksum)s against %(xfer_checksum)s.',
                      {'algo_name': self._hash_algo.name,
                       'checksum': self._expected_checksum,
                       'xfer_checksum': checksum})
            return False
        else:
            LOG.debug('Verifying transfer checksum %(algo_name)s value '
                      '%(checksum)s against %(xfer_checksum)s.',
                      {'algo_name': self._hash_algo.name,
                       'checksum': self._expected_checksum,
                       'xfer_checksum': checksum})
            return True

    @property
    def bytes_transferred(self):
        """Property value to return the number of bytes transferred."""
        return self._bytes_transferred

    @property
    def content_length(self):
        """Property value to return the server indicated length."""
        # If none, there is nothing we can do, the server didn't have
        # a response.
        return self._expected_size


def validate_text_checksum(payload, digest):
    """Compares the checksum of a payload versus the digest.

    The purpose of this method is to take the payload string data,
    and compare it to the digest value of the supplied input. The use
    of this is to validate the the data in cases where we have data
    and need to compare it. Useful in API responses, such as those
    from an OCI Container Registry.

    :param payload: The supplied string with an encode method.
    :param digest: The checksum value in digest form of algorithm:checksum.
    :raises: ImageChecksumError when the response payload does not match the
             supplied digest.
    """
    split_digest = digest.split(':')
    checksum_algo = split_digest[0]
    checksum = split_digest[1]
    hasher = hashlib.new(checksum_algo)
    hasher.update(payload.encode())
    if hasher.hexdigest() != checksum:
        # Mismatch, something is wrong.
        raise exception.ImageChecksumError()



# Copyright 2016 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import functools
import random
import time

from ironic.common import exception
from ironic.common.i18n import _


class Timer(object):
    """A timer decorator and context manager.

    This metric type times the decorated method or code running inside the
    context manager, and emits the time as the metric value. It is bound to
    this MetricLogger.  For example::

      from ironic.common import metrics_utils

      METRICS = metrics_utils.get_metrics_logger()

      @METRICS.timer('foo')
      def foo(bar, baz):
          print bar, baz

      with METRICS.timer('foo'):
          do_something()
    """
    def __init__(self, metrics, name):
        """Init the decorator / context manager.

        :param metrics: The metric logger
        :param name: The metric name
        """
        if not isinstance(name, str):
            raise TypeError(_("The metric name is expected to be a string. "
                            "Value is %s") % name)
        self.metrics = metrics
        self.name = name
        self._start = None

    def __call__(self, f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            start = _time()
            result = f(*args, **kwargs)
            duration = _time() - start

            # Log the timing data (in ms)
            self.metrics.send_timer(self.metrics.get_metric_name(self.name),
                                    duration * 1000)
            return result
        return wrapped

    def __enter__(self):
        self._start = _time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        duration = _time() - self._start
        # Log the timing data (in ms)
        self.metrics.send_timer(self.metrics.get_metric_name(self.name),
                                duration * 1000)


class Counter(object):
    """A counter decorator and context manager.

    This metric type increments a counter every time the decorated method or
    context manager is executed. It is bound to this MetricLogger. For
    example::

      from ironic.common import metrics_utils

      METRICS = metrics_utils.get_metrics_logger()

      @METRICS.counter('foo')
      def foo(bar, baz):
          print bar, baz

      with METRICS.counter('foo'):
          do_something()
    """
    def __init__(self, metrics, name, sample_rate):
        """Init the decorator / context manager.

        :param metrics: The metric logger
        :param name: The metric name
        :param sample_rate: Probabilistic rate at which the values will be sent
        """
        if not isinstance(name, str):
            raise TypeError(_("The metric name is expected to be a string. "
                            "Value is %s") % name)

        if (sample_rate is not None
                and (sample_rate < 0.0 or sample_rate > 1.0)):
            msg = _("sample_rate is set to %s. Value must be None "
                    "or in the interval [0.0, 1.0]") % sample_rate
            raise ValueError(msg)

        self.metrics = metrics
        self.name = name
        self.sample_rate = sample_rate

    def __call__(self, f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            self.metrics.send_counter(
                self.metrics.get_metric_name(self.name),
                1, sample_rate=self.sample_rate)

            result = f(*args, **kwargs)

            return result
        return wrapped

    def __enter__(self):
        self.metrics.send_counter(self.metrics.get_metric_name(self.name),
                                  1, sample_rate=self.sample_rate)

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class Gauge(object):
    """A gauge decorator.

    This metric type returns the value of the decorated method as a metric
    every time the method is executed. It is bound to this MetricLogger. For
    example::

      from ironic.common import metrics_utils

      METRICS = metrics_utils.get_metrics_logger()

      @METRICS.gauge('foo')
      def add_foo(bar, baz):
          return (bar + baz)
    """
    def __init__(self, metrics, name):
        """Init the decorator / context manager.

        :param metrics: The metric logger
        :param name: The metric name
        """
        if not isinstance(name, str):
            raise TypeError(_("The metric name is expected to be a string. "
                            "Value is %s") % name)
        self.metrics = metrics
        self.name = name

    def __call__(self, f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            result = f(*args, **kwargs)
            self.metrics.send_gauge(self.metrics.get_metric_name(self.name),
                                    result)

            return result
        return wrapped


def _time():
    """Wraps time.time() for simpler testing."""
    return time.time()


class MetricLogger(object, metaclass=abc.ABCMeta):
    """Abstract class representing a metrics logger.

    A MetricLogger sends data to a backend (noop or statsd).
    The data can be a gauge, a counter, or a timer.

    The data sent to the backend is composed of:
      - a full metric name
      - a numeric value

    The format of the full metric name is:
        _prefix<delim>name
    where:
        - _prefix: [global_prefix<delim>][uuid<delim>][host_name<delim>]prefix
        - name: the name of this metric
        - <delim>: the delimiter. Default is '.'
    """

    def __init__(self, prefix='', delimiter='.'):
        """Init a MetricLogger.

        :param prefix: Prefix for this metric logger. This string will prefix
            all metric names.
        :param delimiter: Delimiter used to generate the full metric name.
        """
        self._prefix = prefix
        self._delimiter = delimiter

    def get_metric_name(self, name):
        """Get the full metric name.

        The format of the full metric name is:
           _prefix<delim>name
        where:
           - _prefix: [global_prefix<delim>][uuid<delim>][host_name<delim>]
             prefix
           - name: the name of this metric
           - <delim>: the delimiter. Default is '.'


        :param name: The metric name.
        :return: The full metric name, with logger prefix, as a string.
        """
        if not self._prefix:
            return name
        return self._delimiter.join([self._prefix, name])

    def send_gauge(self, name, value):
        """Send gauge metric data.

        Gauges are simple values.
        The backend will set the value of gauge 'name' to 'value'.

        :param name: Metric name
        :param value: Metric numeric value that will be sent to the backend
        """
        self._gauge(name, value)

    def send_counter(self, name, value, sample_rate=None):
        """Send counter metric data.

        Counters are used to count how many times an event occurred.
        The backend will increment the counter 'name' by the value 'value'.

        Optionally, specify sample_rate in the interval [0.0, 1.0] to
        sample data probabilistically where::

            P(send metric data) = sample_rate

        If sample_rate is None, then always send metric data, but do not
        have the backend send sample rate information (if supported).

        :param name: Metric name
        :param value: Metric numeric value that will be sent to the backend
        :param sample_rate: Probabilistic rate at which the values will be
            sent. Value must be None or in the interval [0.0, 1.0].
        """
        if (sample_rate is None or random.random() < sample_rate):
            return self._counter(name, value,
                                 sample_rate=sample_rate)

    def send_timer(self, name, value):
        """Send timer data.

        Timers are used to measure how long it took to do something.

        :param m_name: Metric name
        :param m_value: Metric numeric value that will be sent to the backend
        """
        self._timer(name, value)

    def timer(self, name):
        return Timer(self, name)

    def counter(self, name, sample_rate=None):
        return Counter(self, name, sample_rate)

    def gauge(self, name):
        return Gauge(self, name)

    @abc.abstractmethod
    def _gauge(self, name, value):
        """Abstract method for backends to implement gauge behavior."""

    @abc.abstractmethod
    def _counter(self, name, value, sample_rate=None):
        """Abstract method for backends to implement counter behavior."""

    @abc.abstractmethod
    def _timer(self, name, value):
        """Abstract method for backends to implement timer behavior."""

    def get_metrics_data(self):
        """Return the metrics collection, if available."""
        raise exception.MetricsNotSupported()


class NoopMetricLogger(MetricLogger):
    """Noop metric logger that throws away all metric data."""
    def _gauge(self, name, value):
        pass

    def _counter(self, name, value, sample_rate=None):
        pass

    def _timer(self, m_name, value):
        pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from ironic.common import boot_devices
from ironic.common import exception
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conf import CONF
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules import inspect_utils
from ironic.drivers.modules.inspector import agent as inspector
from ironic.drivers.modules.inspector import interface as common
from ironic.drivers import utils as driver_utils
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


@mock.patch.object(inspect_utils, 'create_ports_if_not_exist', autospec=True)
class InspectHardwareTestCase(db_base.DbTestCase):
    def setUp(self):
        super().setUp()
        self.node = obj_utils.create_test_node(self.context,
                                               inspect_interface='agent')
        self.iface = inspector.AgentInspect()
        self.task = mock.MagicMock(spec=task_manager.TaskManager)
        self.task.context = self.context
        self.task.shared = False
        self.task.node = self.node
        self.task.driver = mock.Mock(
            spec=['boot', 'network', 'inspect', 'power', 'management'],
            inspect=self.iface)
        self.driver = self.task.driver

    def test_unmanaged_ok(self, mock_create_ports_if_not_exist):
        CONF.set_override('require_managed_boot', False, group='inspector')
        self.driver.boot.validate_inspection.side_effect = (
            exception.UnsupportedDriverExtension(''))
        self.assertEqual(states.INSPECTWAIT,
                         self.iface.inspect_hardware(self.task))
        mock_create_ports_if_not_exist.assert_called_once_with(self.task)
        self.assertFalse(self.driver.boot.prepare_ramdisk.called)
        self.assertFalse(self.driver.network.add_inspection_network.called)
        self.driver.management.set_boot_device.assert_called_once_with(
            self.task, device=boot_devices.PXE, persistent=False)
        self.driver.power.set_power_state.assert_has_calls([
            mock.call(self.task, states.POWER_OFF, timeout=None),
            mock.call(self.task, states.POWER_ON, timeout=None),
        ])
        self.assertFalse(self.driver.network.remove_inspection_network.called)
        self.assertFalse(self.driver.boot.clean_up_ramdisk.called)

    def test_unmanaged_disable_power_off(self, mock_create_ports_if_not_exist):
        CONF.set_override('require_managed_boot', False, group='inspector')
        self.driver.boot.validate_inspection.side_effect = (
            exception.UnsupportedDriverExtension(''))
        self.node.disable_power_off = True
        self.assertEqual(states.INSPECTWAIT,
                         self.iface.inspect_hardware(self.task))
        mock_create_ports_if_not_exist.assert_called_once_with(self.task)
        self.assertFalse(self.driver.boot.prepare_ramdisk.called)
        self.assertFalse(self.driver.network.add_inspection_network.called)
        self.driver.management.set_boot_device.assert_called_once_with(
            self.task, device=boot_devices.PXE, persistent=False)
        self.driver.power.reboot.assert_called_once_with(
            self.task, timeout=None)
        self.driver.power.set_power_state.assert_not_called()
        self.assertFalse(self.driver.network.remove_inspection_network.called)
        self.assertFalse(self.driver.boot.clean_up_ramdisk.called)

    def test_unmanaged_disallowed(self, mock_create_ports_if_not_exist):
        self.driver.boot.validate_inspection.side_effect = (
            exception.UnsupportedDriverExtension(''))
        self.assertRaises(exception.UnsupportedDriverExtension,
                          self.iface.inspect_hardware, self.task)

    @mock.patch.object(deploy_utils, 'get_ironic_api_url', autospec=True)
    def test_managed_ok(self, mock_get_url, mock_create_ports_if_not_exist):
        endpoint = 'http://192.169.0.42:6385/v1'
        mock_get_url.return_value = endpoint
        self.assertEqual(states.INSPECTWAIT,
                         self.iface.inspect_hardware(self.task))
        self.driver.boot.prepare_ramdisk.assert_called_once_with(
            self.task, ramdisk_params={
                'ipa-inspection-callback-url':
                f'{endpoint}/continue_inspection',
            })
        self.driver.network.add_inspection_network.assert_called_once_with(
            self.task)
        self.driver.power.set_power_state.assert_has_calls([
            mock.call(self.task, states.POWER_OFF, timeout=None),
            mock.call(self.task, states.POWER_ON, timeout=None),
        ])
        self.assertFalse(self.driver.network.remove_inspection_network.called)
        self.assertFalse(self.driver.boot.clean_up_ramdisk.called)

    @mock.patch.object(deploy_utils, 'get_ironic_api_url', autospec=True)
    def test_managed_unversion_url(self, mock_get_url,
                                   mock_create_ports_if_not_exist):
        endpoint = 'http://192.169.0.42:6385/'
        mock_get_url.return_value = endpoint
        self.assertEqual(states.INSPECTWAIT,
                         self.iface.inspect_hardware(self.task))
        mock_create_ports_if_not_exist.assert_called_once_with(self.task)
        self.driver.boot.prepare_ramdisk.assert_called_once_with(
            self.task, ramdisk_params={
                'ipa-inspection-callback-url':
                f'{endpoint}v1/continue_inspection',
            })
        self.driver.network.add_inspection_network.assert_called_once_with(
            self.task)
        self.driver.power.set_power_state.assert_has_calls([
            mock.call(self.task, states.POWER_OFF, timeout=None),
            mock.call(self.task, states.POWER_ON, timeout=None),
        ])
        self.assertFalse(self.driver.network.remove_inspection_network.called)
        self.assertFalse(self.driver.boot.clean_up_ramdisk.called)

    @mock.patch.object(deploy_utils, 'get_ironic_api_url', autospec=True)
    def test_managed_disable_power_off(self, mock_get_url,
                                       mock_create_ports_if_not_exist):
        endpoint = 'http://192.169.0.42:6385/v1'
        mock_get_url.return_value = endpoint
        self.node.disable_power_off = True
        self.assertEqual(states.INSPECTWAIT,
                         self.iface.inspect_hardware(self.task))
        self.driver.boot.prepare_ramdisk.assert_called_once_with(
            self.task, ramdisk_params={
                'ipa-inspection-callback-url':
                f'{endpoint}/continue_inspection',
            })
        self.driver.network.add_inspection_network.assert_called_once_with(
            self.task)
        self.driver.power.reboot.assert_called_once_with(
            self.task, timeout=None)
        self.driver.power.set_power_state.assert_not_called()
        self.assertFalse(self.driver.network.remove_inspection_network.called)
        self.assertFalse(self.driver.boot.clean_up_ramdisk.called)


@mock.patch.object(driver_utils, 'store_ramdisk_logs', autospec=True)
@mock.patch.object(common, 'tear_down_managed_boot', autospec=True)
@mock.patch.object(inspect_utils, 'run_inspection_hooks', autospec=True)
class ContinueInspectionTestCase(db_base.DbTestCase):
    def setUp(self):
        super().setUp()
        CONF.set_override('enabled_inspect_interfaces',
                          ['agent', 'no-inspect'])
        self.config(hooks='ramdisk-error,architecture,validate-interfaces,'
                          'ports',
                    group='inspector')
        self.node = obj_utils.create_test_node(
            self.context,
            inspect_interface='agent',
            provision_state=states.INSPECTING)
        self.iface = inspector.AgentInspect()
        self.plugin_data = {"logs": "abcd"}

    def test(self, mock_inspection_hooks, mock_tear_down, mock_ramdisk_logs):
        mock_tear_down.return_value = None
        with task_manager.acquire(self.context, self.node.id) as task:
            result = self.iface.continue_inspection(
                task, mock.sentinel.inventory, self.plugin_data)
            mock_inspection_hooks.assert_called_once_with(
                task, mock.sentinel.inventory, self.plugin_data,
                self.iface.hooks, inspector._store_logs)
            mock_tear_down.assert_called_once_with(task, always_power_off=True)
            self.assertEqual(states.INSPECTING, task.node.provision_state)

        self.assertIsNone(result)
        mock_ramdisk_logs.assert_not_called()

    def test_store_logs(self,
                        mock_inspection_hooks,
                        mock_tear_down,
                        mock_ramdisk_logs):
        CONF.set_override('deploy_logs_collect', 'always', group='agent')
        mock_tear_down.return_value = None
        with task_manager.acquire(self.context, self.node.id) as task:
            result = self.iface.continue_inspection(
                task, mock.sentinel.inventory, self.plugin_data)
            mock_inspection_hooks.assert_called_once_with(
                task, mock.sentinel.inventory, self.plugin_data,
                self.iface.hooks, inspector._store_logs)
            mock_tear_down.assert_called_once_with(task, always_power_off=True)
            mock_ramdisk_logs.assert_called_once_with(
                mock.ANY, "abcd", label="inspect")
            self.assertEqual(states.INSPECTING, task.node.provision_state)

        self.assertIsNone(result)


@mock.patch.object(common, 'tear_down_managed_boot', autospec=True)
class AbortInspectionTestCase(db_base.DbTestCase):
    def setUp(self):
        super().setUp()
        CONF.set_override('enabled_inspect_interfaces',
                          ['agent', 'no-inspect'])
        self.node = obj_utils.create_test_node(
            self.context,
            driver_internal_info={
                'lookup_bmc_addresses': [42],
            },
            inspect_interface='agent',
            provision_state=states.INSPECTWAIT)
        self.iface = inspector.AgentInspect()

    def test_success(self, mock_tear_down):
        mock_tear_down.return_value = None
        with task_manager.acquire(self.context, self.node.id) as task:
            self.iface.abort(task)
            mock_tear_down.assert_called_once_with(task, always_power_off=True)

        self.node.refresh()
        # Keep the state - it will be changed on the conductor level
        self.assertEqual(states.INSPECTWAIT, self.node.provision_state)
        self.assertNotIn('lookup_bmc_addresses',
                         self.node.driver_internal_info)

    def test_cleanup_failed(self, mock_tear_down):
        mock_tear_down.return_value = ['boom']
        with task_manager.acquire(self.context, self.node.id) as task:
            self.iface.abort(task)
            mock_tear_down.assert_called_once_with(task, always_power_off=True)

        self.node.refresh()
        self.assertEqual(states.INSPECTFAIL, self.node.provision_state)
        self.assertIn('boom', self.node.last_error)
        self.assertNotIn('lookup_bmc_addresses',
                         self.node.driver_internal_info)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

from ironic.common import exception
from ironic.conductor import task_manager
from ironic.conf import CONF
from ironic.drivers.modules.inspector.hooks import raid_device \
    as raid_device_hook
from ironic.objects.node_inventory import NodeInventory
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


class RaidDeviceTestCase(db_base.DbTestCase):
    def setUp(self):
        super().setUp()
        CONF.set_override('enabled_inspect_interfaces',
                          ['agent', 'no-inspect'])
        self.node = obj_utils.create_test_node(self.context,
                                               inspect_interface='agent')
        self.inventory_1 = {'disks': [{'name': '/dev/sda', 'serial': '1111'},
                                      {'name': '/dev/sdb', 'serial': '2222'}]}
        self.inventory_2 = {'disks': [{'name': '/dev/sdb', 'serial': '2222'},
                                      {'name': '/dev/sdc', 'serial': '3333'}]}
        self.plugin_data = {'plugin_data': 'fake-plugin-data'}

    def test_root_device_already_set(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            self.node.properties = {'root_device': 'any'}
            raid_device_hook.RaidDeviceHook().__call__(task,
                                                       self.inventory_1,
                                                       self.plugin_data)
            self.assertEqual(self.node.properties.get('root_device'), 'any')

    def test_no_serials(self):
        self.inventory_1['disks'][0]['serial'] = None
        with task_manager.acquire(self.context, self.node.id) as task:
            raid_device_hook.RaidDeviceHook().__call__(task,
                                                       self.inventory_1,
                                                       self.plugin_data)
            self.node.refresh()
            self.assertIsNone(self.node.properties.get('root_device'))

    @mock.patch.object(NodeInventory, 'get_by_node_id', autospec=True)
    def test_no_previous_inventory(self, mock_get_by_node_id):
        with task_manager.acquire(self.context, self.node.id) as task:
            mock_get_by_node_id.side_effect = exception.NodeInventoryNotFound()
            raid_device_hook.RaidDeviceHook().__call__(task,
                                                       self.inventory_1,
                                                       self.plugin_data)
            self.node.refresh()
            self.assertIsNone(self.node.properties.get('root_device'))

    @mock.patch.object(NodeInventory, 'get_by_node_id', autospec=True)
    def test_no_new_root_devices(self, mock_get_by_node_id):
        with task_manager.acquire(self.context, self.node.id) as task:

            mock_get_by_node_id.return_value = NodeInventory(
                task.context, id=1, node_id=self.node.id,
                inventory_data=self.inventory_1, plugin_data=self.plugin_data)

            raid_device_hook.RaidDeviceHook().__call__(task,
                                                       self.inventory_1,
                                                       self.plugin_data)
            self.node.refresh()
            result = self.node.properties.get('root_device')
            self.assertIsNone(result)

    @mock.patch.object(NodeInventory, 'get_by_node_id', autospec=True)
    def test_root_device_found(self, mock_get_by_node_id):
        with task_manager.acquire(self.context, self.node.id) as task:
            mock_get_by_node_id.return_value = NodeInventory(
                task.context, id=1, node_id=self.node.id,
                inventory_data=self.inventory_1, plugin_data=self.plugin_data)
            raid_device_hook.RaidDeviceHook().__call__(task,
                                                       self.inventory_2,
                                                       self.plugin_data)
            self.node.refresh()
            result = self.node.properties.get('root_device')
            self.assertEqual(result, {'serial': '3333'})

    @mock.patch.object(NodeInventory, 'get_by_node_id', autospec=True)
    def test_multiple_new_root_devices(self, mock_get_by_node_id):
        with task_manager.acquire(self.context, self.node.id) as task:
            mock_get_by_node_id.return_value = NodeInventory(
                task.context, id=1, node_id=self.node.id,
                inventory_data=self.inventory_1, plugin_data=self.plugin_data)
            self.inventory_2['disks'].append({'name': '/dev/sdd',
                                              'serial': '4444'})
            raid_device_hook.RaidDeviceHook().__call__(task,
                                                       self.inventory_2,
                                                       self.plugin_data)
            self.node.refresh()
            result = self.node.properties.get('root_device')
            self.assertIsNone(result)



# Copyright 2014 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import oslo_messaging as messaging
from oslo_messaging.rpc import dispatcher
from osprofiler import profiler

from ironic.common import context as ironic_context
from ironic.common import exception
from ironic.conf import CONF


TRANSPORT = None
NOTIFICATION_TRANSPORT = None
SENSORS_NOTIFIER = None
VERSIONED_NOTIFIER = None

ALLOWED_EXMODS = [
    exception.__name__,
]
EXTRA_EXMODS = []
GLOBAL_MANAGER = None

MANAGER_TOPIC = 'ironic.conductor_manager'


def init(conf):
    global TRANSPORT, NOTIFICATION_TRANSPORT
    global SENSORS_NOTIFIER, VERSIONED_NOTIFIER
    exmods = get_allowed_exmods()
    TRANSPORT = messaging.get_rpc_transport(conf,
                                            allowed_remote_exmods=exmods)
    NOTIFICATION_TRANSPORT = messaging.get_notification_transport(
        conf,
        allowed_remote_exmods=exmods)

    serializer = RequestContextSerializer(messaging.JsonPayloadSerializer())
    SENSORS_NOTIFIER = messaging.Notifier(NOTIFICATION_TRANSPORT,
                                          serializer=serializer)
    if conf.notification_level is None:
        VERSIONED_NOTIFIER = messaging.Notifier(NOTIFICATION_TRANSPORT,
                                                serializer=serializer,
                                                driver='noop')
    else:
        VERSIONED_NOTIFIER = messaging.Notifier(
            NOTIFICATION_TRANSPORT,
            serializer=serializer,
            topics=CONF.versioned_notifications_topics)


def cleanup():
    global TRANSPORT, NOTIFICATION_TRANSPORT
    global SENSORS_NOTIFIER, VERSIONED_NOTIFIER
    assert TRANSPORT is not None
    assert NOTIFICATION_TRANSPORT is not None
    assert SENSORS_NOTIFIER is not None
    assert VERSIONED_NOTIFIER is not None
    TRANSPORT.cleanup()
    NOTIFICATION_TRANSPORT.cleanup()
    TRANSPORT = NOTIFICATION_TRANSPORT = None
    SENSORS_NOTIFIER = VERSIONED_NOTIFIER = None


def set_defaults(control_exchange):
    messaging.set_transport_defaults(control_exchange)


def get_allowed_exmods():
    return ALLOWED_EXMODS + EXTRA_EXMODS


class RequestContextSerializer(messaging.Serializer):

    def __init__(self, base):
        self._base = base

    def serialize_entity(self, context, entity):
        if not self._base:
            return entity
        return self._base.serialize_entity(context, entity)

    def deserialize_entity(self, context, entity):
        if not self._base:
            return entity
        return self._base.deserialize_entity(context, entity)

    def serialize_context(self, context):
        _context = context.to_dict()
        prof = profiler.get()
        if prof:
            trace_info = {
                "hmac_key": prof.hmac_key,
                "base_id": prof.get_base_id(),
                "parent_id": prof.get_id()
            }
            _context.update({"trace_info": trace_info})
        return _context

    def deserialize_context(self, context):
        trace_info = context.pop("trace_info", None)
        if trace_info:
            profiler.init(**trace_info)
        return ironic_context.RequestContext.from_dict(context)


def get_transport_url(url_str=None):
    return messaging.TransportURL.parse(CONF, url_str)


def get_client(target, version_cap=None, serializer=None):
    assert TRANSPORT is not None
    serializer = RequestContextSerializer(serializer)
    return messaging.get_rpc_client(
        TRANSPORT, target, version_cap=version_cap,
        serializer=serializer)


def get_server(target, endpoints, serializer=None):
    assert TRANSPORT is not None
    serializer = RequestContextSerializer(serializer)
    access_policy = dispatcher.DefaultRPCAccessPolicy
    return messaging.get_rpc_server(TRANSPORT,
                                    target,
                                    endpoints,
                                    executor='eventlet',
                                    serializer=serializer,
                                    access_policy=access_policy)


def get_sensors_notifier(service=None, host=None, publisher_id=None):
    assert SENSORS_NOTIFIER is not None
    if not publisher_id:
        publisher_id = "%s.%s" % (service, host or CONF.host)
    return SENSORS_NOTIFIER.prepare(publisher_id=publisher_id)


def get_versioned_notifier(publisher_id=None):
    assert VERSIONED_NOTIFIER is not None
    assert publisher_id is not None
    return VERSIONED_NOTIFIER.prepare(publisher_id=publisher_id)


def set_global_manager(manager):
    global GLOBAL_MANAGER
    if GLOBAL_MANAGER is not None and manager is not None:
        raise RuntimeError("An attempt to set a global manager twice")
    GLOBAL_MANAGER = manager



#
# Copyright 2022 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_log import log as logging
from oslo_utils import uuidutils

from ironic.conf import CONF
from ironic.dhcp import base

LOG = logging.getLogger(__name__)


class DnsmasqDHCPApi(base.BaseDHCP):
    """API for managing host specific Dnsmasq configuration."""

    def update_port_dhcp_opts(self, port_id, dhcp_options, context=None):
        pass

    def update_dhcp_opts(self, task, options, vifs=None):
        """Send or update the DHCP BOOT options for this node.

        :param task: A TaskManager instance.
        :param options: this will be a list of dicts, e.g.

                        ::

                         [{'opt_name': '67',
                           'opt_value': 'pxelinux.0',
                           'ip_version': 4},
                          {'opt_name': '66',
                           'opt_value': '123.123.123.456',
                           'ip_version': 4}]
        :param vifs: Ignored argument
        """
        node = task.node
        macs = set(self._pxe_enabled_macs(task.ports))

        tag = node.driver_internal_info.get('dnsmasq_tag')
        if not tag:
            tag = uuidutils.generate_uuid()
            node.set_driver_internal_info('dnsmasq_tag', tag)
            node.save()

        option_entries = []

        for option in options:
            try:
                option_entries.append(
                    f'tag:{tag},{option["opt_name"]},{option["opt_value"]}')
            except KeyError as missing:
                LOG.warning('Ignoring option %(opt)s for node %(node)s: '
                            'missing %(missing)s',
                            {'opt': option, 'node': node.uuid,
                             'missing': missing})

        opt_file = self._opt_file_path(node)
        LOG.debug('Writing DHCP options for node %(node)s to %(dest)s: '
                  '%(opts)s', {'node': node.uuid, 'dest': opt_file,
                               'opts': '; '.join(option_entries)})
        with open(opt_file, 'w') as f:
            f.write('\n'.join(option_entries) + '\n')

        for mac in macs:
            # Tag each address with the unique uuid scoped to
            # this node and DHCP transaction
            host_file = self._host_file_path(mac)
            entry = f'{mac},set:{tag},set:ironic'
            LOG.debug('Writing DHCP host file for node %(node)s to %(dest)s: '
                      '%(entry)s', {'node': node.uuid, 'dest': host_file,
                                    'entry': entry})
            with open(host_file, 'w') as f:
                f.write(entry + '\n')

    def _opt_file_path(self, node):
        return os.path.join(CONF.dnsmasq.dhcp_optsdir,
                            'ironic-{}.conf'.format(node.uuid))

    def _host_file_path(self, mac):
        return os.path.join(CONF.dnsmasq.dhcp_hostsdir,
                            'ironic-{}.conf'.format(mac))

    def _pxe_enabled_macs(self, ports):
        for port in ports:
            if port.pxe_enabled:
                yield port.address

    def get_ip_addresses(self, task):
        """Get IP addresses for all ports/portgroups in `task`.

        :param task: a TaskManager instance.
        :returns: List of IP addresses associated with
                  task's ports/portgroups.
        """
        lease_path = CONF.dnsmasq.dhcp_leasefile
        macs = set(self._pxe_enabled_macs(task.ports))
        addresses = []
        with open(lease_path, 'r') as f:
            for line in f.readlines():
                lease = line.split()
                if lease[1] in macs:
                    addresses.append(lease[2])
        LOG.debug('Found addresses for %s: %s',
                  task.node.uuid, ', '.join(addresses))
        return addresses

    def clean_dhcp_opts(self, task):
        """Clean up the DHCP BOOT options for the host in `task`.

        :param task: A TaskManager instance.

        :raises: FailedToCleanDHCPOpts
        """

        node = task.node
        # Discard this unique tag
        node.del_driver_internal_info('dnsmasq_tag')
        node.save()

        # Changing the host rule to ignore will be picked up by dnsmasq
        # without requiring a SIGHUP. When the mac address is active again
        # this file will be replaced with one that applies a new unique tag.
        macs = set(self._pxe_enabled_macs(task.ports))
        for mac in macs:
            host_file = self._host_file_path(mac)
            entry = f'{mac},ignore'
            LOG.debug('Writing DHCP host file for node %(node)s to %(dest)s: '
                      '%(entry)s', {'node': node.uuid, 'dest': host_file,
                                    'entry': entry})
            with open(host_file, 'w') as f:
                f.write(entry + '\n')

        # Deleting the file containing dhcp-option won't remove the rules from
        # dnsmasq but no requests will be tagged with the dnsmasq_tag uuid so
        # these rules will not apply.
        opt_file = self._opt_file_path(node)
        if os.path.exists(opt_file):
            LOG.debug('Removing DHCP options file for node %(node)s at '
                      '%(dest)s', {'node': node.uuid, 'dest': opt_file})
            os.remove(opt_file)

    def supports_ipxe_tag(self):
        """Whether the provider will correctly apply the 'ipxe' tag.

        When iPXE makes a DHCP request, does this provider support adding
        the tag `ipxe` or `ipxe6` (for IPv6). When the provider returns True,
        options can be added which filter on these tags.

        The `dnsmasq` provider sets this to True on the assumption that the
        following is included in the dnsmasq.conf:

        dhcp-match=set:ipxe,175

        :returns: True
        """
        return True



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base code for inspection hooks support."""

import abc

from oslo_config import cfg
from oslo_log import log


CONF = cfg.CONF
LOG = log.getLogger(__name__)


class InspectionHook(metaclass=abc.ABCMeta):  # pragma: no cover
    """Abstract base class for inspection hooks."""

    dependencies = []
    """An ordered list of hooks that must be enabled before this one.

    The items here should be entry point names, not classes.
    """

    def preprocess(self, task, inventory, plugin_data):
        """Hook to run before the main inspection data processing.

        This hook is run even before sanity checks.

        :param task: A TaskManager instance.
        :param inventory: Hardware inventory information sent by the ramdisk.
                          Must not be modified by the hook.
        :param plugin_data: Plugin data sent by the ramdisk. May be modified by
                            the hook.
        :returns: nothing.
        """

    def __call__(self, task, inventory, plugin_data):
        """Hook to run to process inspection data (before Ironic node update).

        This hook is run after node is found and ports are created,
        just before the node is updated with the data.

        :param task: A TaskManager instance.
        :param inventory: Hardware inventory information sent by the ramdisk.
                          Must not be modified by the hook.
        :param plugin_data: Plugin data sent by the ramdisk. May be modified by
                            the hook.
        :returns: nothing.
        """



# Copyright 2012 OpenStack Foundation
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from oslo_log import log
from oslo_serialization import jsonutils
from oslo_utils import timeutils
from oslo_utils import uuidutils

from ironic.common import exception
from ironic.common import keystone
from ironic.conf import CONF

_IMAGE_ATTRIBUTES = ['size', 'disk_format', 'owner',
                     'container_format', 'checksum', 'id',
                     'name', 'created_at', 'updated_at',
                     'deleted_at', 'deleted', 'status',
                     'min_disk', 'min_ram', 'tags', 'visibility',
                     'protected', 'file', 'schema', 'os_hash_algo',
                     'os_hash_value']


LOG = log.getLogger(__name__)
_GLANCE_SESSION = None


def _extract_attributes(image):
    output = {}
    # copies attributes from the openstacksdk Image object
    # to a dictionary
    for attr in _IMAGE_ATTRIBUTES:
        output[attr] = getattr(image, attr, None)

    # copy the properties over to start so that image properties
    # are not nested in another properties key
    output['properties'] = getattr(image, 'properties', {})
    output['schema'] = image.schema

    # attributes already copied so we copy the rest into properties
    copied = set(_IMAGE_ATTRIBUTES)
    copied.add('properties')

    for image_property in set(image) - copied:
        # previously with glanceclient only set things
        # were on the object that came back but the SDK
        # defines every possible attribute so we need to filter
        if image[image_property]:
            output['properties'][image_property] = image[image_property]

    return output


def _convert_timestamps_to_datetimes(image_meta):
    """Convert timestamps to datetime objects

    Returns image metadata with timestamp fields converted to naive UTC
    datetime objects.
    """
    for attr in ['created_at', 'updated_at', 'deleted_at']:
        if image_meta.get(attr):
            image_meta[attr] = timeutils.normalize_time(
                timeutils.parse_isotime(image_meta[attr]))
    return image_meta


_CONVERT_PROPS = ('block_device_mapping', 'mappings')


def _convert(metadata):
    metadata = copy.deepcopy(metadata)
    properties = metadata.get('properties')
    if properties:
        for attr in _CONVERT_PROPS:
            if attr in properties:
                prop = properties[attr]
                if isinstance(prop, str):
                    properties[attr] = jsonutils.loads(prop)
    return metadata


def parse_image_id(image_href):
    """Parse an image id from image href.

    :param image_href: href of an image
    :returns: image id parsed from image_href

    :raises InvalidImageRef: when input image href is invalid
    """
    image_href = str(image_href)
    if uuidutils.is_uuid_like(image_href):
        image_id = image_href
    elif image_href.startswith('glance://'):
        image_id = image_href.split('/')[-1]
        if not uuidutils.is_uuid_like(image_id):
            raise exception.InvalidImageRef(image_href=image_href)
    else:
        raise exception.InvalidImageRef(image_href=image_href)
    return image_id


def translate_from_glance(image):
    image_meta = _extract_attributes(image)
    image_meta = _convert_timestamps_to_datetimes(image_meta)
    image_meta = _convert(image_meta)
    return image_meta


def is_image_available(context, image):
    """Check image availability.

    This check is needed in case Nova and Glance are deployed
    without authentication turned on.
    """
    # NOTE: Any support for private/shared images in Ironic requires a secure
    # way for ironic to know the original requester:
    #  - If we trust node[instance_info][project_id], we are susceptible to a
    #    node.owner stealing another project's private image by lying in
    #    instance_info.
    #  - As of 2025.1, the project_id attached to the auth context at this
    #    point is more likely to be the nova-computes service user rather
    #    than the original requester. This is a missing feature from the
    #    Ironic/Nova virt driver.

    auth_token = getattr(context, 'auth_token', None)
    conductor_project_id = get_conductor_project_id()
    image_visibility = getattr(image, 'visibility', None)
    image_owner = getattr(image, 'owner', None)
    image_id = getattr(image, 'id', 'unknown')
    is_admin = 'admin' in getattr(context, 'roles', [])
    project = getattr(context, 'project', 'unknown')
    # If an auth token is present and the config allows access via auth token,
    #  allow image access.
    if CONF.allow_image_access_via_auth_token and auth_token:
        # We return true here since we want the *user* request context to
        # be able to be used.
        return True
    # If the image visibility is public or community, allow access.
    if image_visibility in ['public', 'community']:
        return True
    # If the user is an admin and the config allows ignoring project checks for
    #  admin tasks, allow access.
    if is_admin and CONF.ignore_project_check_for_admin_tasks:
        return True
    # If the image is private and the owner is the conductor project,
    #  allow access.
    if image_visibility == 'private' and image_owner == conductor_project_id:
        return True

    LOG.info(
        'Access to %s owned by %s denied to requester %s',
        image_id, image_owner, project
    )
    return False


def is_image_active(image):
    """Check the image status.

    This check is needed in case the Glance image is stuck in queued status
    or pending_delete.
    """
    return str(getattr(image, 'status', None)) == "active"


def is_glance_image(image_href):
    if not isinstance(image_href, str):
        return False
    return (image_href.startswith('glance://')
            or uuidutils.is_uuid_like(image_href))


def get_conductor_project_id():
    global _GLANCE_SESSION
    if not _GLANCE_SESSION:
        _GLANCE_SESSION = keystone.get_session('glance')
    session = _GLANCE_SESSION
    service_auth = keystone.get_auth('glance')

    try:
        if service_auth and hasattr(service_auth, 'get_project_id'):
            return service_auth.get_project_id(session)
        elif hasattr(session, 'get_project_id') and session.auth:
            return session.get_project_id()
    except Exception as e:
        LOG.debug("Error getting conductor project ID: %s", str(e))
    return None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions for rules."""

from collections import abc

from ironic.common.i18n import _


class ShallowMaskList(abc.MutableSequence):
    """A proxy list to maintain original list and applies masking on the fly.

    This class implements the MutableSequence ABC to provide a complete
    list-like interface while handling sensitive data masking consistently
    with ShallowMaskDict.
    """
    def __init__(self, original_list, sensitive_fields=None,
                 mask_enabled=True):
        self._original = original_list
        self._sensitive_fields = sensitive_fields or []
        self._mask_enabled = mask_enabled

    def _mask_value(self, value):
        """Apply masking to value on demand."""
        if isinstance(value, dict):
            return ShallowMaskDict(
                value,
                sensitive_fields=self._sensitive_fields,
                mask_enabled=self._mask_enabled
            )
        elif isinstance(value, list):
            return ShallowMaskList(
                value,
                sensitive_fields=self._sensitive_fields,
                mask_enabled=self._mask_enabled
            )
        return value

    def __getitem__(self, index):
        value = self._original[index]
        return self._mask_value(value)

    def __setitem__(self, index, value):
        self._original[index] = value

    def __delitem__(self, index):
        del self._original[index]

    def __len__(self):
        return len(self._original)

    def insert(self, index, value):
        self._original.insert(index, value)

    def copy(self):
        return ShallowMaskList(
            self._original.copy(),
            sensitive_fields=self._sensitive_fields,
            mask_enabled=self._mask_enabled
        )

    def set_mask_enabled(self, mask_enabled):
        self._mask_enabled = mask_enabled

    def __repr__(self):
        items = [repr(self._mask_value(item)) for item in self._original]
        return "[%s]" % ", ".join(items)


class ShallowMaskDict(abc.MutableMapping):
    """Dictionary wrapper to mask sensitive fields on the fly.

    This class implements the MutableMapping ABC to provide a complete
    dict-like interface while masking sensitive fields when accessed.
    """
    def __init__(self, data, sensitive_fields=None, mask_enabled=True):
        self._data = data
        self._sensitive_fields = sensitive_fields or []
        self._mask_enabled = mask_enabled

    def _mask_value(self, key, value):
        if self._mask_enabled and key in self._sensitive_fields:
            return '***'

        if isinstance(value, dict):
            return ShallowMaskDict(
                value,
                sensitive_fields=self._sensitive_fields,
                mask_enabled=self._mask_enabled
            )
        elif isinstance(value, list):
            return ShallowMaskList(
                value,
                sensitive_fields=self._sensitive_fields,
                mask_enabled=self._mask_enabled
            )
        return value

    def __getitem__(self, key):
        value = self._data[key]
        return self._mask_value(key, value)

    def __setitem__(self, key, value):
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def copy(self):
        return ShallowMaskDict(
            self._data.copy(),
            sensitive_fields=self._sensitive_fields,
            mask_enabled=self._mask_enabled,
        )

    def set_mask_enabled(self, mask_enabled):
        self._mask_enabled = mask_enabled

    def __repr__(self):
        items = ["%s: %s" % (repr(k), repr(self._mask_value(k, v)))
                 for k, v in self._data.items()]
        return "{%s}" % ", ".join(items)


def parse_inverted_operator(op):
    """Handle inverted operators.

    Parses a logical condition operator to determine if it has been negated
    using a single leading exclamation mark ('!'). Ensures only one
    exclamation mark is allowed.

    Example Usage:
        parse_inverted_operator("!eq")   # Returns ("eq", True)
        parse_inverted_operator(" eq ")  # Returns ("eq", False)
        parse_inverted_operator("!!eq")  # Raises ValueError

    raises: ValueError: If multiple exclamation marks are present
    returns: A tuple containing the cleaned operator and a
            boolean indicating whether negation was applied.
    """
    op = op.strip()

    if op.count('!') > 1:
        msg = _("Multiple exclamation marks are not allowed. "
                "To apply the invert of an operation, simply add an "
                "exclamation mark (with an optional space) before "
                "the operator, e.g. eq - !eq.")
        raise ValueError(msg)

    is_inverted = op.startswith('!')
    op = op.lstrip('!').strip()
    return op, is_inverted


def normalize_path(path):
    """Convert a path (dot or slash notation) to a list of path parts"""
    if '/' in path:
        parts = path.strip('/').split('/')
    else:
        parts = path.split('.')
    return parts



# Copyright (c) 2018 NEC, Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_config import cfg
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import utils
from oslo_upgradecheck import common_checks
from oslo_upgradecheck import upgradecheck
import sqlalchemy

from ironic.cmd import dbsync
from ironic.common import driver_factory
from ironic.common.i18n import _
from ironic.common import policy  # noqa importing to load policy config.
import ironic.conf
from ironic.db import api as db_api

CONF = ironic.conf.CONF


class Checks(upgradecheck.UpgradeCommands):

    """Upgrade checks for the ironic-status upgrade check command

    Upgrade checks should be added as separate methods in this class
    and added to _upgrade_checks tuple.
    """

    def _check_obj_versions(self):
        """Check that the DB versions of objects are compatible.

        Checks that the object versions are compatible with this
        release of ironic. It does this by comparing the objects'
        .version field in the database, with the expected versions
        of these objects.
        """
        try:
            # NOTE(TheJulia): Seems an exception is raised by sqlalchemy
            # when a table is missing, so lets catch it, since it is fatal.
            msg = dbsync.DBCommand().check_obj_versions(
                ignore_missing_tables=True)
        except sqlalchemy.exc.NoSuchTableError as e:
            msg = ('Database table missing. Please ensure you have '
                   'updated the database schema. Not Found: %s' % e)
            return upgradecheck.Result(upgradecheck.Code.FAILURE, details=msg)

        if not msg:
            return upgradecheck.Result(upgradecheck.Code.SUCCESS)
        else:
            return upgradecheck.Result(upgradecheck.Code.FAILURE, details=msg)

    def _check_db_indexes(self):
        """Check if indexes exist on heavily used columns.

        Checks the database to see if indexes exist on heavily used columns
        and provide guidance of action that can be taken to improve ironic
        database performance.
        """
        engine = enginefacade.reader.get_engine()

        indexes = [
            ('nodes', 'reservation_idx'),
            ('nodes', 'driver_idx'),
            ('nodes', 'provision_state_idx'),
            ('nodes', 'conductor_group_idx'),
            ('nodes', 'resource_class_idx'),
            ('nodes', 'reservation_idx'),
            ('nodes', 'owner_idx'),
            ('nodes', 'lessee_idx'),
        ]
        missing_indexes = []
        for table, idx in indexes:
            if not utils.index_exists(engine, table, idx):
                missing_indexes.append(idx)

        if missing_indexes:
            idx_list = ', '.join(missing_indexes)
            msg = ('Indexes missing for ideal database performance. Please '
                   'consult https://docs.openstack.org/ironic/latest/admin/'
                   'tuning.html for information on indexes. Missing: %s'
                   % idx_list)
            return upgradecheck.Result(upgradecheck.Code.WARNING, details=msg)
        else:
            return upgradecheck.Result(upgradecheck.Code.SUCCESS)

    def _check_allocations_table(self):
        msg = None
        engine = enginefacade.reader.get_engine()
        if 'mysql' != str(engine.url.get_backend_name()):
            # This test only applies to mysql and database schema
            # selection.
            return upgradecheck.Result(upgradecheck.Code.SUCCESS)
        with engine.connect() as conn, conn.begin():
            res = conn.execute(
                sqlalchemy.text("show create table allocations"))
        results = str(res.all()).lower()
        if 'utf8' not in results:
            msg = ('The Allocations table is is not using UTF8 encoding. '
                   'This is corrected in later versions of Ironic, where '
                   'the table character set schema is automatically '
                   'migrated. Continued use of a non-UTF8 character '
                   'set may produce unexpected results.')

        if 'innodb' not in results:
            warning = ('The engine used by MySQL for the allocations '
                       'table is not the intended engine for the Ironic '
                       'database tables to use. This may have been a result '
                       'of an error with the table creation schema. This '
                       'may require Database Administrator intervention '
                       'and downtime to dump, modify the table engine to '
                       'utilize InnoDB, and reload the allocations table to '
                       'utilize the InnoDB engine.')
            if msg:
                msg = msg + ' Additionally: ' + warning
            else:
                msg = warning

        if msg:
            return upgradecheck.Result(upgradecheck.Code.WARNING, details=msg)
        else:
            return upgradecheck.Result(upgradecheck.Code.SUCCESS)

    def _check_hardware_types_interfaces(self):
        try:
            hw_types = driver_factory.hardware_types()
        except Exception as exc:
            # NOTE(dtantsur): if the hardware types failed to load, we cannot
            # validate the hardware interfaces, so returning early.
            msg = f"Some hardware types cannot be loaded: {exc}"
            return upgradecheck.Result(upgradecheck.Code.FAILURE, details=msg)

        try:
            ifaces = driver_factory.all_interfaces()
        except Exception as exc:
            msg = f"Some hardware interfaces cannot be loaded: {exc}"
            return upgradecheck.Result(upgradecheck.Code.FAILURE, details=msg)

        warnings = []
        for name, obj in hw_types.items():
            if not obj.supported:
                warnings.append(f"Hardware type {name} is deprecated or not "
                                "supported")

        for iface_type, iface_dict in ifaces.items():
            iface_type = iface_type.capitalize()
            for name, obj in iface_dict.items():
                if not obj.supported:
                    warnings.append(f"{iface_type} interface {name} is "
                                    "deprecated or not supported")

        dbapi = db_api.get_instance()
        for node in dbapi.get_node_list():
            if node.driver not in hw_types:
                warnings.append(f"Node {node.uuid} uses an unknown driver "
                                f"{node.driver}")
            for iface_type, iface_dict in ifaces.items():
                value = getattr(node, f"{iface_type}_interface")
                # NOTE(dtantsur): the interface value can be empty if a new
                # interface type has just been added, and nodes have not been
                # updated yet.
                if value and value not in iface_dict:
                    warnings.append(f"Node {node.uuid} uses an unknown "
                                    f"{iface_type} interface {value}")

        if warnings:
            msg = ". ".join(warnings)
            return upgradecheck.Result(upgradecheck.Code.WARNING, details=msg)
        else:
            return upgradecheck.Result(upgradecheck.Code.SUCCESS)

    # A tuple of check tuples of (<name of check>, <check function>).
    # The name of the check will be used in the output of this command.
    # The check function takes no arguments and returns an
    # oslo_upgradecheck.upgradecheck.Result object with the appropriate
    # oslo_upgradecheck.upgradecheck.Code and details set. If the
    # check function hits warnings or failures then those should be stored
    # in the returned Result's "details" attribute. The
    # summary will be rolled up at the end of the check() method.
    _upgrade_checks = (
        (_('Object versions'), _check_obj_versions),
        (_('Database Index Status'), _check_db_indexes),
        (_('Allocations Name Field Length Check'),
         _check_allocations_table),
        # Victoria -> Wallaby migration
        (_('Policy File JSON to YAML Migration'),
         (common_checks.check_policy_json, {'conf': CONF})),
        (_('Hardware Types and Interfaces Check'),
         _check_hardware_types_interfaces),
    )


def main():
    return upgradecheck.main(
        cfg.CONF, project='ironic', upgrade_command=Checks())


if __name__ == '__main__':
    sys.exit(main())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic.common import states

# These flags tell the conductor that we're rebooting inside or after a step.

CLEANING_REBOOT = "cleaning_reboot"
DEPLOYMENT_REBOOT = "deployment_reboot"
SERVICING_REBOOT = "servicing_reboot"

# These flags tell the conductor whether the currently running step should be
# skipped after the agent heartbeats again. Setting them to false causes
# the conductor to re-enter the previously running step after a reboot.

SKIP_CURRENT_CLEAN_STEP = "skip_current_clean_step"
SKIP_CURRENT_DEPLOY_STEP = "skip_current_deploy_step"
SKIP_CURRENT_SERVICE_STEP = "skip_current_service_step"

# These flags tell the conductor that something else (most likely, a periodic
# task in some hardware interface) is polling the step for completion.

CLEANING_POLLING = "cleaning_polling"
DEPLOYMENT_POLLING = "deployment_polling"
SERVICING_POLLING = "servicing_polling"

_ALL_FLAGS = [CLEANING_REBOOT, DEPLOYMENT_REBOOT, SERVICING_REBOOT,
              SKIP_CURRENT_CLEAN_STEP, SKIP_CURRENT_DEPLOY_STEP,
              SKIP_CURRENT_SERVICE_STEP,
              CLEANING_POLLING, DEPLOYMENT_POLLING, SERVICING_POLLING]


def get_return_state(node):
    """Returns state based on operation being invoked.

    :param node: an ironic node object.
    :returns: states.CLEANWAIT if cleaning operation in progress,
              or states.DEPLOYWAIT if deploy operation in progress,
              or states.SERVICEWAIT if servicing in progress.
    """
    # FIXME(dtantsur): this distinction is rather useless, create a new
    # constant to use for all step types?
    if node.clean_step:
        return states.CLEANWAIT
    elif node.service_step:
        return states.SERVICEWAIT
    else:
        # TODO(dtantsur): ideally, check for node.deploy_step and raise
        # something if this function is called without any step field set.
        # Unfortunately, a lot of unit tests rely on exactly this.
        return states.DEPLOYWAIT


def _check_agent_token_prior_to_agent_reboot(node):
    """Removes the agent token if it was not pregenerated.

    Removal of the agent token in cases where it is not pregenerated
    is a vital action prior to rebooting the agent, as without doing
    so the agent will be unable to establish communication with
    the ironic API after the reboot. Effectively locking itself out
    as in cases where the value is not pregenerated, it is not
    already included in the payload and must be generated again
    upon lookup.

    :param node: The Node object.
    """
    if not node.driver_internal_info.get('agent_secret_token_pregenerated',
                                         False):
        node.del_driver_internal_info('agent_secret_token')


def _step_type(node, step_type):
    if step_type:
        return step_type
    if node.clean_step:
        return 'clean'
    elif node.service_step:
        return 'service'
    else:
        return 'deploy'


def set_node_flags(node, reboot=None, skip_current_step=None, polling=None,
                   step_type=None):
    """Sets appropriate reboot flags in driver_internal_info based on operation

    :param node: an ironic node object.
    :param reboot: Boolean value to set for node's driver_internal_info flag
        cleaning_reboot, servicing_reboot or deployment_reboot based on the
        operation in progress. If it is None, corresponding reboot flag is
        not set in node's driver_internal_info.
    :param skip_current_step: Boolean value to set for node's
        driver_internal_info flag skip_current_clean_step,
        skip_current_service_step or skip_current_deploy_step based on the
        operation in progress. If it is None, corresponding skip step flag is
        not set in node's driver_internal_info.
    :param polling: Boolean value to set for node's driver_internal_info flag
        deployment_polling, servicing_polling or cleaning_polling. If it is
        None, the corresponding polling flag is not set in the node's
        driver_internal_info. A polling flag is otherwise deleted from
        the node's driver_internal_info.
    :param step_type: The type of steps to process: 'clean', 'service'
        or 'deploy'. If None, detected from the node.
    """
    step_type = _step_type(node, step_type)
    if step_type == 'clean':
        reboot_field = CLEANING_REBOOT
        skip_field = SKIP_CURRENT_CLEAN_STEP
        polling_field = CLEANING_POLLING
    elif step_type == 'service':
        reboot_field = SERVICING_REBOOT
        skip_field = SKIP_CURRENT_SERVICE_STEP
        polling_field = SERVICING_POLLING
    else:
        reboot_field = DEPLOYMENT_REBOOT
        skip_field = SKIP_CURRENT_DEPLOY_STEP
        polling_field = DEPLOYMENT_POLLING

    if reboot is not None:
        node.set_driver_internal_info(reboot_field, reboot)
        if reboot:
            # If rebooting, we must ensure that we check and remove
            # an agent token if necessary.
            _check_agent_token_prior_to_agent_reboot(node)
    if skip_current_step is not None:
        node.set_driver_internal_info(skip_field, skip_current_step)
    if polling is not None:
        node.set_driver_internal_info(polling_field, polling)
    else:
        # Always unset the value *if* previously set.
        node.del_driver_internal_info(polling_field)
    node.save()


def remove_node_flags(node):
    """Remove all flags for the node.

    :param node: A Node object
    """
    for flag in _ALL_FLAGS:
        node.del_driver_internal_info(flag)


def prepare_node_for_next_step(node, step_type=None):
    """Remove the flags responsible for the next step.

    Cleans the polling and the skip-next step flags.

    :param node: A Node object
    :param step_type: The type of steps to process: 'clean', 'service'
        or 'deploy'. If None, detected from the node.
    :returns: The last value of the skip-next flag.
    """
    step_type = _step_type(node, step_type)
    skip_current_step = node.del_driver_internal_info(
        'skip_current_%s_step' % step_type, True)
    if step_type == 'clean':
        node.del_driver_internal_info(CLEANING_POLLING)
    elif step_type == 'service':
        node.del_driver_internal_info(SERVICING_POLLING)
    else:
        node.del_driver_internal_info(DEPLOYMENT_POLLING)
    return skip_current_step



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import time
from unittest import mock

from oslo_config import cfg
import oslo_messaging
from oslo_service import service as base_service
from oslo_utils import timeutils

from ironic.common import context
from ironic.common import rpc
from ironic.common import service as ironic_service
from ironic.conductor import manager
from ironic.conductor import rpc_service
from ironic.objects import base as objects_base
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils

CONF = cfg.CONF


@mock.patch.object(base_service.Service, '__init__', lambda *_, **__: None)
class TestRPCService(db_base.DbTestCase):

    def setUp(self):
        super(TestRPCService, self).setUp()
        host = "fake_host"
        mgr_module = "ironic.conductor.manager"
        mgr_class = "ConductorManager"
        self.rpc_svc = rpc_service.RPCService(host, mgr_module, mgr_class)
        # register oslo_service DEFAULT config options
        ironic_service.process_launcher()
        self.rpc_svc.manager.dbapi = self.dbapi

    @mock.patch.object(manager.ConductorManager, 'prepare_host', autospec=True)
    @mock.patch.object(oslo_messaging, 'Target', autospec=True)
    @mock.patch.object(objects_base, 'IronicObjectSerializer', autospec=True)
    @mock.patch.object(rpc, 'get_server', autospec=True)
    @mock.patch.object(manager.ConductorManager, 'init_host', autospec=True)
    @mock.patch.object(context, 'get_admin_context', autospec=True)
    def test_start(self, mock_ctx, mock_init_method,
                   mock_rpc, mock_ios, mock_target, mock_prepare_method):
        mock_rpc.return_value.start = mock.MagicMock()
        self.rpc_svc.handle_signal = mock.MagicMock()
        self.assertFalse(self.rpc_svc._started)
        self.assertFalse(self.rpc_svc._failure)
        self.rpc_svc.start()
        mock_ctx.assert_called_once_with()
        mock_target.assert_called_once_with(topic=self.rpc_svc.topic,
                                            server="fake_host")
        mock_ios.assert_called_once_with(is_server=True)
        mock_prepare_method.assert_called_once_with(self.rpc_svc.manager)
        mock_init_method.assert_called_once_with(self.rpc_svc.manager,
                                                 mock_ctx.return_value)
        self.assertIs(rpc.GLOBAL_MANAGER, self.rpc_svc.manager)
        self.assertTrue(self.rpc_svc._started)
        self.assertFalse(self.rpc_svc._failure)
        self.rpc_svc.wait_for_start()  # should be no-op

    @mock.patch.object(manager.ConductorManager, 'prepare_host', autospec=True)
    @mock.patch.object(oslo_messaging, 'Target', autospec=True)
    @mock.patch.object(objects_base, 'IronicObjectSerializer', autospec=True)
    @mock.patch.object(rpc, 'get_server', autospec=True)
    @mock.patch.object(manager.ConductorManager, 'init_host', autospec=True)
    @mock.patch.object(context, 'get_admin_context', autospec=True)
    def test_start_no_rpc(self, mock_ctx, mock_init_method,
                          mock_rpc, mock_ios, mock_target,
                          mock_prepare_method):
        CONF.set_override('rpc_transport', 'none')
        self.rpc_svc.start()

        self.assertIsNone(self.rpc_svc.rpcserver)
        mock_ctx.assert_called_once_with()
        mock_target.assert_not_called()
        mock_rpc.assert_not_called()
        mock_ios.assert_called_once_with(is_server=True)
        mock_prepare_method.assert_called_once_with(self.rpc_svc.manager)
        mock_init_method.assert_called_once_with(self.rpc_svc.manager,
                                                 mock_ctx.return_value)
        self.assertIs(rpc.GLOBAL_MANAGER, self.rpc_svc.manager)

    @mock.patch.object(manager.ConductorManager, 'prepare_host', autospec=True)
    @mock.patch.object(oslo_messaging, 'Target', autospec=True)
    @mock.patch.object(objects_base, 'IronicObjectSerializer', autospec=True)
    @mock.patch.object(rpc, 'get_server', autospec=True)
    @mock.patch.object(manager.ConductorManager, 'init_host', autospec=True)
    @mock.patch.object(context, 'get_admin_context', autospec=True)
    def test_start_failure(self, mock_ctx, mock_init_method, mock_rpc,
                           mock_ios, mock_target, mock_prepare_method):
        mock_rpc.return_value.start = mock.MagicMock()
        self.rpc_svc.handle_signal = mock.MagicMock()
        mock_init_method.side_effect = RuntimeError("boom")
        self.assertFalse(self.rpc_svc._started)
        self.assertFalse(self.rpc_svc._failure)
        self.assertRaises(RuntimeError, self.rpc_svc.start)
        mock_ctx.assert_called_once_with()
        mock_target.assert_called_once_with(topic=self.rpc_svc.topic,
                                            server="fake_host")
        mock_ios.assert_called_once_with(is_server=True)
        mock_prepare_method.assert_called_once_with(self.rpc_svc.manager)
        mock_init_method.assert_called_once_with(self.rpc_svc.manager,
                                                 mock_ctx.return_value)
        self.assertIsNone(rpc.GLOBAL_MANAGER)
        self.assertFalse(self.rpc_svc._started)
        self.assertIn("boom", self.rpc_svc._failure)
        self.assertRaises(SystemExit, self.rpc_svc.wait_for_start)

    @mock.patch.object(timeutils, 'utcnow', autospec=True)
    @mock.patch.object(time, 'sleep', autospec=True)
    def test_stop_instant(self, mock_sleep, mock_utcnow):
        # del_host returns instantly
        mock_utcnow.return_value = datetime.datetime(2023, 2, 2, 21, 10, 0)
        conductor1 = db_utils.get_test_conductor(hostname='fake_host')
        with mock.patch.object(self.dbapi, 'get_online_conductors',
                               autospec=True) as mock_cond_list:
            mock_cond_list.return_value = [conductor1]
            with mock.patch.object(self.dbapi, 'get_nodeinfo_list',
                                   autospec=True) as mock_nodeinfo_list:
                mock_nodeinfo_list.return_value = []
                self.rpc_svc.stop()

        # single conductor so exit immediately without waiting
        mock_sleep.assert_not_called()

    @mock.patch.object(timeutils, 'utcnow', autospec=True)
    @mock.patch.object(time, 'sleep', autospec=True)
    def test_stop_after_full_reset_interval(self, mock_sleep, mock_utcnow):
        # del_host returns instantly
        mock_utcnow.return_value = datetime.datetime(2023, 2, 2, 21, 10, 0)
        conductor1 = db_utils.get_test_conductor(hostname='fake_host')
        conductor2 = db_utils.get_test_conductor(hostname='other_fake_host')
        with mock.patch.object(self.dbapi, 'get_online_conductors',
                               autospec=True) as mock_cond_list:
            # multiple conductors, so wait for hash_ring_reset_interval
            mock_cond_list.return_value = [conductor1, conductor2]
            with mock.patch.object(self.dbapi, 'get_nodeinfo_list',
                                   autospec=True) as mock_nodeinfo_list:
                mock_nodeinfo_list.return_value = []
                self.rpc_svc.stop()
                mock_nodeinfo_list.assert_called_once()

        # wait the total CONF.hash_ring_reset_interval 15 seconds
        mock_sleep.assert_has_calls([mock.call(15)])

    @mock.patch.object(timeutils, 'utcnow', autospec=True)
    @mock.patch.object(time, 'sleep', autospec=True)
    def test_stop_after_remaining_interval(self, mock_sleep, mock_utcnow):
        mock_utcnow.return_value = datetime.datetime(2023, 2, 2, 21, 10, 0)
        conductor1 = db_utils.get_test_conductor(hostname='fake_host')
        conductor2 = db_utils.get_test_conductor(hostname='other_fake_host')

        # del_host returns after 5 seconds
        mock_utcnow.side_effect = [
            datetime.datetime(2023, 2, 2, 21, 10, 0),
            datetime.datetime(2023, 2, 2, 21, 10, 5),
        ]
        with mock.patch.object(self.dbapi, 'get_online_conductors',
                               autospec=True) as mock_cond_list:
            # multiple conductors, so wait for hash_ring_reset_interval
            mock_cond_list.return_value = [conductor1, conductor2]
            with mock.patch.object(self.dbapi, 'get_nodeinfo_list',
                                   autospec=True) as mock_nodeinfo_list:
                mock_nodeinfo_list.return_value = []
                self.rpc_svc.stop()
                mock_nodeinfo_list.assert_called_once()

        # wait the remaining 10 seconds
        mock_sleep.assert_has_calls([mock.call(10)])

    @mock.patch.object(timeutils, 'utcnow', autospec=True)
    @mock.patch.object(time, 'sleep', autospec=True)
    def test_stop_slow(self, mock_sleep, mock_utcnow):
        mock_utcnow.return_value = datetime.datetime(2023, 2, 2, 21, 10, 0)
        conductor1 = db_utils.get_test_conductor(hostname='fake_host')
        conductor2 = db_utils.get_test_conductor(hostname='other_fake_host')

        # del_host returns after 16 seconds
        mock_utcnow.side_effect = [
            datetime.datetime(2023, 2, 2, 21, 10, 0),
            datetime.datetime(2023, 2, 2, 21, 10, 16),
        ]
        with mock.patch.object(self.dbapi, 'get_online_conductors',
                               autospec=True) as mock_cond_list:
            # multiple conductors, so wait for hash_ring_reset_interval
            mock_cond_list.return_value = [conductor1, conductor2]
            with mock.patch.object(self.dbapi, 'get_nodeinfo_list',
                                   autospec=True) as mock_nodeinfo_list:
                mock_nodeinfo_list.return_value = []
                self.rpc_svc.stop()
                mock_nodeinfo_list.assert_called_once()

        # no wait required, CONF.hash_ring_reset_interval already exceeded
        mock_sleep.assert_not_called()

    @mock.patch.object(timeutils, 'utcnow', autospec=True)
    @mock.patch.object(time, 'sleep', autospec=True)
    def test_stop_has_reserved(self, mock_sleep, mock_utcnow):
        mock_utcnow.return_value = datetime.datetime(2023, 2, 2, 21, 10, 0)
        conductor1 = db_utils.get_test_conductor(hostname='fake_host')
        conductor2 = db_utils.get_test_conductor(hostname='other_fake_host')

        with mock.patch.object(self.dbapi, 'get_online_conductors',
                               autospec=True) as mock_cond_list:
            # multiple conductors, so wait for hash_ring_reset_interval
            mock_cond_list.return_value = [conductor1, conductor2]
            with mock.patch.object(self.dbapi, 'get_nodeinfo_list',
                                   autospec=True) as mock_nodeinfo_list:
                # 3 calls to manager has_reserved until all reservation locks
                # are released
                mock_nodeinfo_list.side_effect = [['a', 'b'], ['a'], []]
                self.rpc_svc.stop()
                self.assertEqual(3, mock_nodeinfo_list.call_count)

        # wait the remaining 15 seconds, then wait until has_reserved
        # returns False
        mock_sleep.assert_has_calls(
            [mock.call(15), mock.call(1), mock.call(1)])

    @mock.patch.object(timeutils, 'utcnow', autospec=True)
    @mock.patch.object(time, 'sleep', autospec=True)
    def test_drain_has_reserved(self, mock_sleep, mock_utcnow):
        mock_utcnow.return_value = datetime.datetime(2023, 2, 2, 21, 10, 0)
        conductor1 = db_utils.get_test_conductor(hostname='fake_host')
        conductor2 = db_utils.get_test_conductor(hostname='other_fake_host')

        with mock.patch.object(self.dbapi, 'get_online_conductors',
                               autospec=True) as mock_cond_list:
            # multiple conductors, so wait for hash_ring_reset_interval
            mock_cond_list.return_value = [conductor1, conductor2]
            with mock.patch.object(self.dbapi, 'get_nodeinfo_list',
                                   autospec=True) as mock_nodeinfo_list:
                # 3 calls to manager has_reserved until all reservation locks
                # are released
                mock_nodeinfo_list.side_effect = [['a', 'b'], ['a'], []]
                self.rpc_svc._handle_drain(None, None)
                self.assertEqual(3, mock_nodeinfo_list.call_count)

        # wait the remaining 15 seconds, then wait until has_reserved
        # returns False
        mock_sleep.assert_has_calls(
            [mock.call(15), mock.call(1), mock.call(1)])

    @mock.patch.object(timeutils, 'utcnow', autospec=True)
    def test_shutdown_timeout_reached(self, mock_utcnow):

        initial_time = datetime.datetime(2023, 2, 2, 21, 10, 0)
        before_graceful = initial_time + datetime.timedelta(seconds=30)
        after_graceful = initial_time + datetime.timedelta(seconds=90)
        before_drain = initial_time + datetime.timedelta(seconds=1700)
        after_drain = initial_time + datetime.timedelta(seconds=1900)

        mock_utcnow.return_value = before_graceful
        self.assertFalse(self.rpc_svc._shutdown_timeout_reached(initial_time))

        mock_utcnow.return_value = after_graceful
        self.assertTrue(self.rpc_svc._shutdown_timeout_reached(initial_time))

        self.rpc_svc.draining = True
        self.assertFalse(self.rpc_svc._shutdown_timeout_reached(initial_time))

        mock_utcnow.return_value = before_drain
        self.assertFalse(self.rpc_svc._shutdown_timeout_reached(initial_time))

        mock_utcnow.return_value = after_drain
        self.assertTrue(self.rpc_svc._shutdown_timeout_reached(initial_time))

        CONF.set_override('drain_shutdown_timeout', 0)
        self.assertFalse(self.rpc_svc._shutdown_timeout_reached(initial_time))



# -*- encoding: utf-8 -*-
#
# Copyright © 2012 eNovance <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log
try:
    from oslo_reports import guru_meditation_report as gmr
    from oslo_reports import opts as gmr_opts
except ImportError:
    gmr = None
from oslo_service import service

from ironic.common import config
from ironic.common import profiler
from ironic.conf import CONF
from ironic.conf import opts
from ironic import objects
from ironic import version


LOG = log.getLogger(__name__)


def prepare_command(argv=None):
    """Prepare any Ironic command for execution.

    Sets up configuration and logging, registers objects.
    """
    argv = [] if argv is None else argv
    log.register_options(CONF)
    opts.update_opt_defaults()
    config.parse_args(argv)
    # NOTE(vdrok): We need to setup logging after argv was parsed, otherwise
    # it does not properly parse the options from config file and uses defaults
    # from oslo_log
    log.setup(CONF, 'ironic')
    objects.register_all()


def prepare_service(name, argv=None, conf=CONF):
    """Prepare an Ironic service executable.

    In addition to what `prepare_command` does, set up guru meditation
    reporting and profiling.
    """
    prepare_command(argv)

    if gmr is not None:
        gmr_opts.set_defaults(CONF)
        gmr.TextGuruMeditation.setup_autorun(version, conf=CONF)
    else:
        LOG.debug('Guru meditation reporting is disabled '
                  'because oslo.reports is not installed')

    profiler.setup(name, CONF.host)


def process_launcher():
    return service.ProcessLauncher(CONF, restart_method='mutate')


def ensure_rpc_transport(conf=CONF):
    # Only the combined ironic executable can use rpc_transport = none
    if conf.rpc_transport == 'none':
        raise RuntimeError("This service is not designed to work with "
                           "rpc_transport = none. Please use the combined "
                           "ironic executable or another RPC transport.")



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Abstract base class for console container providers.
"""

import abc


class BaseConsoleContainer(object, metaclass=abc.ABCMeta):
    """Base class for console container provider APIs."""

    @abc.abstractmethod
    def start_container(self, task, app_name, app_info):
        """Start a console container for a node.

        Calling this will block until a consumable container host and port can
        be returned.

        :param task: A TaskManager instance.
        :param app_name: Name of app to run in the container
        :param app_info: Dict of app-specific info
        :returns: Tuple of host IP address and published port
        :raises: ConsoleContainerError
        """

    @abc.abstractmethod
    def stop_container(self, task):
        """Stop a console container for a node.

        Any existing running container for this node will be stopped.

        :param task: A TaskManager instance.
        :raises: ConsoleContainerError
        """

    @abc.abstractmethod
    def stop_all_containers(self):
        """Stops all running console containers

        This is run on conductor startup and graceful shutdown to ensure
        no console containers are running.
        :raises: ConsoleContainerError
        """



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Test class for DRAC inspection interface
"""

from unittest import mock

from oslo_utils import units
import sushy

from ironic.common import states
from ironic.conductor import task_manager
from ironic.drivers.modules.drac import inspect as drac_inspect
from ironic.drivers.modules import inspect_utils
from ironic.drivers.modules.redfish import inspect as redfish_inspect
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic.tests.unit.drivers.modules.drac import utils as test_utils
from ironic.tests.unit.objects import utils as obj_utils

INFO_DICT = test_utils.INFO_DICT


class DracRedfishInspectionTestCase(test_utils.BaseDracTest):
    def setUp(self):
        super(DracRedfishInspectionTestCase, self).setUp()
        self.config(enabled_hardware_types=['idrac'],
                    enabled_power_interfaces=['idrac-redfish'],
                    enabled_management_interfaces=['idrac-redfish'],
                    enabled_inspect_interfaces=['idrac-redfish'])
        self.node = obj_utils.create_test_node(
            self.context, driver='idrac',
            driver_info=INFO_DICT)

    def init_system_mock(self, system_mock, **properties):
        system_mock.reset()
        system_mock.boot.mode = 'uefi'
        system_mock.bios.attributes = {
            'PxeDev1EnDis': 'Enabled', 'PxeDev2EnDis': 'Disabled',
            'PxeDev3EnDis': 'Disabled', 'PxeDev4EnDis': 'Disabled',
            'PxeDev1Interface': 'NIC.Integrated.1-1-1',
            'PxeDev2Interface': None, 'PxeDev3Interface': None,
            'PxeDev4Interface': None}

        system_mock.memory_summary.size_gib = 2

        system_mock.processors.summary = '8', 'MIPS'

        system_mock.simple_storage.disks_sizes_bytes = (
            1 * units.Gi, units.Gi * 3, units.Gi * 5)
        system_mock.storage.volumes_sizes_bytes = (
            2 * units.Gi, units.Gi * 4, units.Gi * 6)

        system_mock.ethernet_interfaces.summary = {
            '00:11:22:33:44:55': sushy.STATE_ENABLED,
            '24:6E:96:70:49:00': sushy.STATE_DISABLED}
        member_data = [{
            'description': 'Integrated NIC 1 Port 1 Partition 1',
            'name': 'System Ethernet Interface',
            'full_duplex': False,
            'identity': 'NIC.Integrated.1-1-1',
            'mac_address': '24:6E:96:70:49:00',
            'mtu_size': None,
            'speed_mbps': 0,
            'vlan': None}]
        system_mock.ethernet_interfaces.get_members.return_value = [
            test_utils.dict_to_namedtuple(values=interface)
            for interface in member_data
        ]
        return system_mock

    def test_get_properties(self):
        expected = redfish_utils.COMMON_PROPERTIES
        driver = drac_inspect.DracRedfishInspect()
        self.assertEqual(expected, driver.get_properties())

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__get_pxe_port_macs_with_UEFI_boot_mode(self, mock_get_system):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.boot.mode = 'uefi'
        expected_pxe_mac = ['24:6E:96:70:49:00']
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            pxe_port_macs = task.driver.inspect._get_pxe_port_macs(task)
            self.assertEqual(expected_pxe_mac, pxe_port_macs)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__get_pxe_port_macs_with_BIOS_boot_mode(self, mock_get_system):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.boot.mode = 'bios'
        mock_manager = mock.MagicMock()
        system_mock.managers = [mock_manager]
        mock_manager_oem = mock_manager.get_oem_extension.return_value
        mock_manager_oem.get_pxe_port_macs_bios.return_value = \
            ['24:6E:96:70:49:00']
        expected_pxe_mac = ['24:6E:96:70:49:00']
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            pxe_port_macs = task.driver.inspect._get_pxe_port_macs(task)
            self.assertEqual(expected_pxe_mac, pxe_port_macs)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__get_pxe_port_macs_without_boot_mode(self, mock_get_system):
        system_mock = self.init_system_mock(mock_get_system.return_value)
        system_mock.boot.mode = None
        expected_pxe_mac = []
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            pxe_port_macs = task.driver.inspect._get_pxe_port_macs(task)
            self.assertEqual(expected_pxe_mac, pxe_port_macs)

    @mock.patch.object(redfish_inspect.RedfishInspect, 'inspect_hardware',
                       autospec=True)
    @mock.patch.object(inspect_utils, 'create_ports_if_not_exist',
                       autospec=True)
    def test_inspect_hardware_with_ethernet_interfaces_mac(
            self, mock_create_ports_if_not_exist, mock_inspect_hardware):
        ethernet_interfaces_mac = {'NIC.Integrated.1-1-1':
                                   '24:6E:96:70:49:00'}
        mock_inspect_hardware.return_value = states.MANAGEABLE
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.driver.inspect._get_mac_address = mock.Mock()
            task.driver.inspect._get_mac_address.return_value = \
                ethernet_interfaces_mac
            return_value = task.driver.inspect.inspect_hardware(task)
            self.assertEqual(states.MANAGEABLE, return_value)
            mock_create_ports_if_not_exist.assert_called_once_with(
                task, ['24:6E:96:70:49:00'])

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__get_mac_address_with_ethernet_interfaces(self, mock_get_system):
        self.init_system_mock(mock_get_system.return_value)
        expected_value = {'NIC.Integrated.1-1-1': '24:6E:96:70:49:00'}
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            return_value = task.driver.inspect._get_mac_address(task)
            self.assertEqual(expected_value, return_value)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__get_mac_address_without_ethernet_interfaces(self,
                                                          mock_get_system):
        mock_system = self.init_system_mock(mock_get_system.return_value)
        mock_system.ethernet_interfaces.summary = None
        expected_value = {}
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            return_value = task.driver.inspect._get_mac_address(task)
            self.assertEqual(expected_value, return_value)
